/*--------------------------------------------------------------
/* 			SPSS/Win 11.5 or Higher Macro -- Written by David Marso
/*		 	Robust Variance Estimation in Meta-Regression with Dependent Effect Size Estimates 
/*
/* 			To use, initialize macro with the include statement:
/* 			INCLUDE "[drive][path]RobustMeta.SPS".													*/ 
/*
/* 			Syntax for macro:
/* 			ROBUST STUDYID = varname                                                                                                     
/*                   / EFFSIZE = varname
/*                   / VAREFFS = varname
/*                   / RHO = value 
/*                   / DESIGN = varnames
/*		      	 									/[ WEIGHTS = varname ]
/*                   /[ RESID = option ]
/*                   /[ HWEIGHT = option ]
/*                   /[ PRINT = option ].
/*
/* 			Where STUDYID is the study identifier variable, EFFSIZE is the effect size variable, VAREFFS is the 
/*  		sampling variance of the effect size estimates, RHO is the ICC used to calculate variance components in the 
/*  		random effects model, which must be < 1, and DESIGN is a list of the independent  variables in the model. 
/* 			The /WEIGHTS subcommand can be used to specify user-defined weights. 
/*  		If /WEIGHTS is omitted, the default weights calculations are 1/(k*meanvariance) in the random effects case, 
/*  		and (1/variance) in the hierarchical case.
/* 			The /RESID = T subcommand is used to open a new dataset that stores the fitted residuals from the model. 
/* 			If /RESID is omitted, no residuals will be stored. The /HWEIGHT=T subcommand is used for a hierarchical model 
/*  		weighting scheme when effect size estimates are nested within studies that are nested within clusters. 
/* 			If /PRINT = EXP, the output will print the exponent of the B coefficient (the odds-ratio/risk-ratio).  
/* 			If /PRINT is omitted, the unstandardized regression coefficient is printed.
/*
/* 			Example:
/* 			ROBUST STUDYID = studyid / EFFSIZE = es / VAREFFS = variance / RHO = .70 / DESIGN = male age / RESID =T.      
/*
/*				Description:
/*    ROBUST provides a robust method for estimating standard errors in meta-regression, particularly when 
/*  		there are dependent effects. Dependent effects occur in two basic models:  (1) correlated effects and 
/*  		(2) hierarchical meta-regression. In (1), the dependency arises as a result of correlated estimation errors; 
/*  		for example, a study collects two outcome measures on each participant sample and then summarizes these 
/* 			as two effect size measures. In (2), the dependency arises as a result of correlated parameters; for example, 
/*  		the same research group may publish several studies and there may be elements of these studies that are 
/*  		similar to one another. Importantly, the robust standard error procedure used here does not require the 
/*  		underlying correlation structure to be known; additionally, it works for any weights and can be used to 
/*  		estimate the mean effect size as well as meta-regression models. Finally, note that the procedure also can 
/*  		be used with independent	effects, particularly when distributional assumptions might be violated. 
/*
/*  		For more information on the underlying theory, see the following references.
/*				References:
/*    Hedges, Larry V., Elizabeth Tipton, and Matthew C. Johnson. 2010. Robust variance estimation in 
/*					meta-regression with dependent effect size estimates. Research Synthesis Methods, 1, 39-65.  
/* 				(www.interscience.wiley.com) DOI: 10.1002/jrsm.5
/*    Website for further information: http://www.northwestern.edu/ipr/qcenter/RVE-meta-analysis.html
/*
/* 			Version 2011.03.07
/*			-------------------------------------------------------------- */.
set printback=off.

DEFINE ROBUST ( STUDYID !CHAREND ("/") 
              / EFFSIZE !CHAREND ("/")
              / VAREFFS !CHAREND ("/")
              / RHO     !CHAREND ("/") !DEFAULT ("")
              / DESIGN  !CHAREND ("/") !DEFAULT ("")
              / WEIGHTS !CHAREND ("/") !DEFAULT ("")
              / RESID   !CHAREND ("/") !DEFAULT ("")
              / HWEIGHT !CHAREND ("/") !DEFAULT ("")
              / PRINT   !CHAREND ("/") !DEFAULT (DEF)  ).
PRESERVE.
SET MPRINT OFF .
SET PRINTBACK OFF.

STRING #CMDSYN (A255).
COMPUTE #CMDSYN=CONCAT("ROBUST",
                       " STUDYID=",!QUOTE(!STUDYID),
                      "/ EFFSIZE=",!QUOTE(!EFFSIZE),
                      "/ VAREFFS=",!QUOTE(!VAREFFS)    ).
!IF (!RHO !NE !NULL) !THEN
COMPUTE #CMDSYN=CONCAT(RTRIM(#CMDSYN),"/ RHO=",!QUOTE(!RHO)).
!IFEND
!IF (!DESIGN !NE !NULL) !THEN
COMPUTE #CMDSYN=CONCAT(RTRIM(#CMDSYN),"/ DESIGN=",!QUOTE(!DESIGN)).
!IFEND
!IF (!WEIGHTS !NE !NULL) !THEN
COMPUTE #CMDSYN=CONCAT(RTRIM(#CMDSYN),"/ WEIGHTS=",!QUOTE(!WEIGHTS)).
!IFEND
!IF (!RESID !NE !NULL) !THEN
COMPUTE #CMDSYN=CONCAT(RTRIM(#CMDSYN),"/ RESID=",!QUOTE(!RESID)).
!IFEND
!IF (!HWEIGHT !NE !NULL) !THEN
COMPUTE #CMDSYN=CONCAT(RTRIM(#CMDSYN),"/ HWEIGHT=",!QUOTE(!HWEIGHT)).
!IFEND
!IF (!PRINT !NE DEF) !THEN
COMPUTE #CMDSYN=CONCAT(RTRIM(#CMDSYN),"/ PRINT=",!QUOTE(!PRINT)).
!IFEND

SET WIDTH=255.
DO IF $CASENUM=1.
PRINT / #CMDSYN.
END IF

*** Check parameters **.
!LET !FMT="F10.6"
!LET !ERR="X"

!IF (!studyid !EQ !NULL) !THEN

DO IF $CASENUM=1.
PRINT /"STUDYID parameter is missing.  Please correct this and resubmit command.".
END IF.

!LET !ERR="T"
!IFEND.


!IF (!HWEIGHT !NE !NULL) !THEN
DO IF $CASENUM=1.
PRINT /!QUOTE(!CONCAT("Hierarchical method requested.  Weights will be calculated as 1 / ",!VAREFFS)).

!IF (!WEIGHTS !NE !NULL) !THEN
PRINT / "Specified weights will be ignored".
!IFEND 
END IF.
COMPUTE @WEIGHT@=1/!VAREFFS.
!IFEND


!IF (!effsize !EQ !NULL) !THEN
DO IF $CASENUM=1.
PRINT /"EFFSIZE parameter is missing.  Please correct this and resubmit command." .

END IF.
!LET !ERR="T"
!IFEND.

!IF (!vareffs !EQ !NULL) !THEN
DO IF $CASENUM=1.
PRINT /"VAREFFS parameter is missing.  Please correct this and resubmit command.".

END IF.
!LET !ERR="T"
!IFEND.

!IF (!HWEIGHT !EQ !NULL) !THEN
!IF (!RHO !EQ !NULL) !THEN
DO IF $CASENUM=1.
PRINT /"RHO parameter is missing.  0 will be assumed.".

END IF.
!IFEND
!IFEND

* Select only cases with complete data on all required variables *.
!IF (!ERR !EQ "X") !THEN 
COMPUTE @NMISS@=0.
COMPUTE @NP@=0.
DO REPEAT V=!DESIGN !effsize.
IF MISSING(V) @NMISS@=@NMISS@+1.
COMPUTE @NP@=@NP@+1.
END REPEAT.

SELECT IF @NMISS@=0.
COMPUTE @ORD@=$CASENUM.

** Build contiguous studyID and build K as counter *.

SORT CASES BY !studyid  .
AUTORECODE !studyid / INTO @STID@ .
* Create Counter for K *.
DO IF $CASENUM=1 OR @STID@ <> LAG (@STID@ ).
COMPUTE @K@=1.
COMPUTE @MEFF@=!vareffs.
ELSE .
COMPUTE @K@=LAG(@K@) + 1.
COMPUTE @MEFF@=!vareffs+LAG(@MEFF@).
END IF.

* Now place largest K on top and drag through the studyID * .
SORT CASES BY @STID@ (D) @K@ (D).
DO IF @STID@ <> LAG (@STID@ ) OR $CASENUM=1.
COMPUTE @MEFF@ =@MEFF@ /@K@.
ELSE.
COMPUTE @K@ =LAG(@K@).
COMPUTE @MEFF@ =LAG(@MEFF@).
END IF.

!IF (!HWEIGHT !EQ !NULL) !THEN
!IF (!WEIGHTS !EQ !NULL )!THEN
DO IF $CASENUM=1.
PRINT /!QUOTE("WEIGHTS parameter is missing.  Weights will be calculated as 1/(k*meanvariance) ").
END IF.
COMPUTE @WEIGHT@=1/(@K@*@MEFF@) .
!ELSE
COMPUTE @WEIGHT@=!WEIGHTS.
!IFEND
!IFEND

COMPUTE @S@=SQRT(!vareffs).
DO IF $CASENUM=1.
+  COMPUTE @NN@=@STID@ .
+  COMPUTE @TDF@=@NN@-@NP@.
+  COMPUTE @T_CRIT@=IDF.T(.975,@TDF@).
ELSE.
+  COMPUTE @T_CRIT@=LAG(@T_CRIT@).
END IF.

COMPUTE INTERCEP=1.
SET MXLOOP=20000.
MATRIX.
GET ID     / VARIABLES @STID@             / FILE=* .
GET K      / VARIABLES @K@                / FILE=* .
GET WT     / VARIABLES @WEIGHT@           / FILE=* .
GET VAREFFS /VARIABLES !VAREFFS           / FILE= *.
GET Y      / VARIABLES !effsize           / FILE=* .
GET S      / VARIABLES @S@                / FILE=* .
GET MEANV  / VARIABLES @MEFF@             / FILE=* .
GET TCRIT  / VARIABLES=@T_CRIT@           / FILE * .
GET DES    / VARIABLES INTERCEP !DESIGN   / FILE */ NAMES=Params.

COMPUTE N=CMAX(ID(:,1)).                         /***N <- max(data$study).
COMPUTE NEFS=NROW(ID(:,1)).

COMPUTE P=NCOL(DES)-1.                           /***p <- ncol(X.full)-2 /The -2 here is presumably due to the study k vectors */.
COMPUTE sumXWX   = 0.                            /***sumXWX <- 0 ***/.
COMPUTE sumXWy   = 0.                            /***sumXWy<- 0.
COMPUTE sumXWJWX = 0.                            /***sumXWJWX <- 0.
COMPUTE sumXWVWX = 0.                            /***sumXWVWX <- 0.
COMPUTE sumXWsig = 0.                            /***sumXW.sig.m.v.WX <- 0.

!IF (!HWEIGHT !NE !NULL) !THEN
PRINT /TITLE "Hierachical method assumes Weights are 1/VarEffect Size.".
!IF (!WEIGHTS !NE !NULL) !THEN
PRINT /TITLE "Specified Weights will be ignored".
!IFEND

+  COMPUTE sumXWWX  = 0.
+  COMPUTE sumXJX   = 0.
+  COMPUTE sumXWJJX = 0.
+  COMPUTE sumXJJWX = 0.
+  COMPUTE trsumJJ  = 0.
+  COMPUTE sumXJWX  = 0.
+  COMPUTE sumXWJX  = 0.

COMPUTE numRows=NROW(ID).

COMPUTE curRow=1.
LOOP i= 1 TO N .                                  /***for (i in (1: N)) {.
+  COMPUTE Ki=K(curRow,1).
+  COMPUTE endRow=curRow+Ki-1.
+  COMPUTE Yi=Y(curRow:endRow,1).                 /***Y <- cbind(tab$effectsize).
+  COMPUTE Si=S(curRow:endRow,1).
+  COMPUTE WTi=WT(curRow:endRow,1).
+  COMPUTE Xi=DES(curRow:endRow,:).               /*grab the design information.
+  COMPUTE J=MAKE(Ki,Ki,1.0).                     /* Create a unit matrix of size K*K.
+  COMPUTE sigma = Si*T(Si).                      /***sigma <- (tab$s%*%t(tab$s)).

+  DO IF (Ki > 1).
+     COMPUTE W=MDIAG(WTi).
+     COMPUTE vee=INV(W).
+     COMPUTE SigmV = sigma - vee.
+  ELSE IF (Ki = 1).

+     COMPUTE W=WTI.
+     COMPUTE vee=1/W.
+     COMPUTE SigmV = sigma - vee.
+  END IF.
+  COMPUTE TXW=T(Xi)*W.
+  COMPUTE TTXW=T(TXW).
+  COMPUTE sumXWX = sumXWX + TXW*Xi.                /***sumXWX <- sumXWX + t(X)%*%W%*%X.
+  COMPUTE sumXWy = sumXWy + TXW*Yi.                /***sumXWy <- sumXWy + t(X)%*%W%*%Y.
+  COMPUTE sumXWJWX = sumXWJWX + TXW*J*W*Xi.        /***sumXWJWX <- sumXWJWX + t(X)%*%W%*%J%*%W%*%X.
+  COMPUTE sumXWVWX = sumXWVWX + TXW*vee*W*Xi.       /***sumXWVWX <- sumXWVWX + t(X)%*%W%*%vee%*%W%*%X.
+  COMPUTE sumXWsig = sumXWsig + TXW*SigmV*W*Xi.     /***sumXW.sig.m.v.WX <- sumXW.sig.m.v.WX + t(X)%*%W%*%SigmV%*%W%*%X.
+  COMPUTE curRow=curRow+Ki.
+  COMPUTE JJ=J*J.
+  COMPUTE trsumJJ = trsumJJ + TRACE(JJ).           /**tr.sumJJ <- tr.sumJJ + sum(diag(J%*%J)).
+  COMPUTE sumXWWX  = sumXWWX  + TXW*W*Xi.          /**sumXWWX  <- sumXWWX  + t(X)%*%W%*%W%*%X.
+  COMPUTE sumXJX   = sumXJX   + t(Xi)*J*Xi.        /**sumXJX   <- sumXJX   + t(X)%*%J%*%X.
+  COMPUTE sumXWJJX = sumXWJJX + TXW*JJ*Xi.         /**sumXWJJX <- sumXWJJX + t(X)%*%W%*%J%*%J%*%X.
+  COMPUTE sumXJJWX = sumXJJWX + t(Xi)*JJ*W*Xi.     /**sumXJJWX <- sumXJJWX + t(X)%*%J%*%J%*%W%*%X.
+  COMPUTE sumXJWX  = sumXJWX  + t(Xi)*J*W*Xi.      /**sumXJWX  <- sumXJWX  + t(X)%*%J%*%W%*%X.
+  COMPUTE sumXWJX  = sumXWJX  + TXW *J*Xi.         /**sumXWJX <-  sumXWJX  + t(X)%*%W%*%J%*%X.
END LOOP.                                                            /***}.
                                    

COMPUTE InvSXwX=INV(sumXWX).
COMPUTE b = InvSXwX *sumXWy.                                         /***b <- solve(sumXWX)%*%sumXWy.

COMPUTE X=DES   .                                                    /***X <- data.matrix(X.full[-c(1)]).
COMPUTE pred = X*b.                                                  /***data$pred <- X%*%b.
COMPUTE e = Y - pred.                                                /***data$e <- data$effect.size - data$pred.
COMPUTE W = MDIAG(WT).                                               /***W <- diag(data$weights).
COMPUTE sumW = CSUM(WT).                                             /***sumW <- sum(data$weights).
COMPUTE Qe = t(e)*W*e.                                               /***Qe <-t(data$e)%*%W%*%data$e.
** Calculations for Tau Sq **.
COMPUTE denom  = sumW - Trace(InvSXwX*sumXWJWX).                      /***denom <- sumW - sum(diag(solve(sumXWX)%*%sumXWJWX)).
COMPUTE term1  = (Qe  - N + Trace(InvSXwX *sumXWVWX)) /denom.         /***term1 <- (Qe - N + sum(diag(solve(sumXWX)%*%sumXWVWX))) /denom.
COMPUTE term2  = Trace(InvSXwX *sumXWsig) / denom.                    /***term2 <- (sum(diag(solve(sumXWX)%*%sumXW.sig.m.v.WX ))) / denom.

** Calculations for OmegaSq **.

+  COMPUTE curRow=1.
+  COMPUTE sumEJE=0.
+  COMPUTE TOTALN=0.
+  LOOP i= 1 TO N .                                                     /***for (i in (1: N)) {.
+    COMPUTE Ki=K(curRow,1).
+    COMPUTE endRow=curRow+Ki-1.
+    COMPUTE Ei=E(curRow:endRow,1).                                    /***e <- tab$e.
+    COMPUTE Ki=K(curRow,1).
+  DO IF (Ki > 1).
+     COMPUTE Ji=MAKE(Ki,Ki,1)). 
+  ELSE IF (Ki = 1).
+     COMPUTE Ji=1. 
+  END IF.
+    COMPUTE sumEJE=sumEJE+T(Ei)*Ji*Ei.                                /***sumEJE <- sumEJE + t(e)%*%J%*%e.
+    COMPUTE curRow=curRow+Ki.
+    COMPUTE TOTALN=TOTALN + Ki.
+  END LOOP.

+  COMPUTE sumV = Msum(VAREFFS).                                      /***sumV <- sum(data$vareffsize).
+  COMPUTE W  = INV(MDIAG(VAREFFS)).                                  /***W <- diag(1/data$vareffsize).                                        
+  COMPUTE SUMW=MSUM(W).                                              /***sumW <-sum(W).
+  COMPUTE Qa = sumEJE.                                               /***Qa <- sumEJE.
+  COMPUTE Vi = INV(sumXWX).                                          /***V.i <- solve(sumXWX).

**+ # Qe terms .
+  COMPUTE A = sumW - TRACE(Vi * sumXWJWX).                           /***A<-sumW - sum(diag(V.i%*%sumXWJWX)).
+  COMPUTE B = sumW - TRACE(Vi * sumXWWX).                            /***B<- sumW - sum(diag(V.i%*%sumXWWX)).
+  COMPUTE C = TotalN - (p+1).                                        /***C <- length(data$k) - (p+1).

**+ # Qa terms, Qa = y'(I-H)'J.s(I-H)y, A = J.s - H'J.s-J.sH + H'J.sH .

***a1 <- tr.sumJJ - sum(diag(V.i%*%sumXJJWX)) - sum(diag(V.i%*%sumXWJJX)) + sum(diag(V.i%*%sumXJX%*%V.i%*%sumXWJWX)).
+  COMPUTE  a1 = trsumJJ- TRACE(Vi*sumXJJWX) - TRACE(Vi*sumXWJJX) + TRACE(Vi*sumXJX*Vi*sumXWJWX) .                        

***b1 <- length(data$k) - sum(diag(V.i%*%sumXWJX)) - sum(diag(V.i%*%sumXJWX)) + sum(diag(V.i%*%sumXJX%*%V.i%*%sumXWWX)). 
+  COMPUTE  b1 = TotalN - TRACE(Vi * sumXWJX)- TRACE(Vi * sumXJWX) + TRACE(Vi * sumXJX * Vi * sumXWWX).                

***c1 <- sumV - sum(diag(V.i%*%sumXJX)).
+  COMPUTE  c1 = sumV - TRACE(Vi*sumXJX).                             
**  #est omega2 and tau2 .        .
***omega.sq1 <- ((Qa - c1)*A - (Qe - C)*a1)/(b1*A - B*a1) . 
+  COMPUTE  omegasq1 = ((Qa - c1)*A - (Qe - C)*a1)/(b1*A - B*a1).   
+  COMPUTE omegasq =omegasq1 .
+  DO IF omegasq <=0.
+    COMPUTE omegasq =0.
+  END IF.
+  COMPUTE tausq1=((Qe-C)/A)-omegasq*(B/A).
+  DO IF tausq1<=0.
+    COMPUTE tausq1=0.
+  END IF.

** HIERARCH COMPUTE rwts = 1/(K &*(MEANV + tausq)).                   /***data$r.weights <- 1/(data$k*(data$mean.v + tau.sq).
COMPUTE DiagTerm=omegasq+tausq1.

COMPUTE rwts=VAREFFS.
COMPUTE sumXWXr = 0.                                                 /***sumXWX.r <- 0.
COMPUTE sumXWyr = 0.                                                 /***sumXWy.r<- 0.
COMPUTE curRow=1.
LOOP i=1 TO N.
+  COMPUTE Ki      = K(curRow,1).
+  COMPUTE endRow  = curRow+Ki-1.
+  DO IF (Ki > 1).
+     COMPUTE DTDiag=MDIAG(MAKE(1,Ki,DiagTerm)).
+     COMPUTE Wi    = MDIAG(rwts(curRow:endRow,1)).
+     COMPUTE Wi=Wi + DTDiag.
+     COMPUTE Wi=INV(Wi). 
+  ELSE IF (Ki = 1).
+     COMPUTE Wi    = rwts(curRow:endRow,1)).
+     COMPUTE Wi=1/(Wi + DiagTerm).
+  END IF.
+  COMPUTE Xi      = DES(curRow:endRow,:).
+  COMPUTE Yi      = Y(curRow:endRow,1).
+  COMPUTE TXW     = T(Xi)*Wi.
+  COMPUTE sumXWXr = sumXWXr + TXW*Xi.
+  COMPUTE sumXWyr = sumXWyr + TXW*Yi.
+  COMPUTE curRow  = curRow+Ki.
END LOOP.

COMPUTE br         = INV(sumXWXr)*sumXWyr.                      /***b.r <- solve(sumXWX.r)%*%sumXWy.r .
COMPUTE X          = DES .                                      /***X <- data.matrix(X.full[-c(1)].
COMPUTE predr      = X*br.                                      /***data$pred.r <- X%*%b.r .
COMPUTE er         = Y - predr.                                 /***data$e.r <- cbind(data$effect.size) - data$pred.r.

COMPUTE sumXWWXr   = 0.                                         /***sumXWeeWX.r <- 0.
COMPUTE curRow     = 1.
LOOP i=1 TO N.                                                  /***for (i in (1:N)) {.

+  COMPUTE KI      = K(curRow,1).
+  COMPUTE endRow  = curRow+KI-1.

+  COMPUTE erI     = er(curRow:endRow,1).
+  COMPUTE sighatr = erI*T(erI).                                /***sigma.hat.r <- tab$e.r%*%t(tab$e.r).
                              
+  DO IF (Ki > 1).
+     COMPUTE DTDiag=MDIAG(MAKE(1,Ki,DiagTerm)).
+     COMPUTE Wi    = MDIAG(rwts(curRow:endRow,1)).
+     COMPUTE Wi=Wi + DTDiag.
+     COMPUTE Wr=INV(Wi). 
+  ELSE IF (Ki = 1).
+     COMPUTE Wi    = rwts(curRow:endRow,1)).
+     COMPUTE Wr=1/(Wi + DiagTerm).
+  END IF.
+  COMPUTE X       = DES(curRow:endRow,:).
+  COMPUTE TXW     = T(X)*Wr.
+  COMPUTE sumXWWXr= sumXWWXr + TXW*sighatr*Wr*X.               /***sumXWeeWX.r <- sumXWeeWX.r + t(X)%*%W%*%sigma.hat.r%*%W%*%X.
+  COMPUTE curRow  = curRow+KI.
END LOOP.

!IFEND   /** END OF HWEIGHT /**.

!IF (!HWEIGHT !EQ !NULL) !THEN
** Do Non Hierarchical analysis **.

COMPUTE numRows=NROW(ID).

COMPUTE curRow=1.
LOOP i= 1 TO N .                                  /***for (i in (1: N)) {
+  COMPUTE Ki=K(curRow,1).
+  COMPUTE endRow=curRow+Ki-1.
+  COMPUTE Yi=Y(curRow:endRow,1).                 /***Y <- cbind(tab$effectsize)
+  COMPUTE Si=S(curRow:endRow,1).
+  COMPUTE WTi=WT(curRow:endRow,1).
+  COMPUTE Xi=DES(curRow:endRow,:).               /*grab the design information.
+  COMPUTE J=MAKE(Ki,Ki,1.0).                     /* Create a unit matrix of size K*K.
+  COMPUTE sigma = Si*T(Si).                      /***sigma <- (tab$s%*%t(tab$s)).
+  COMPUTE v=MAKE(1,Ki,Si(1,1)**2).
+  DO IF (Ki > 1).
+     COMPUTE vee=MDIAG(v).
+     COMPUTE W=MDIAG(WTI).
+     COMPUTE SigmV = sigma - vee.
+  ELSE IF (Ki = 1).
+     COMPUTE vee=v.
+     COMPUTE W=WTI.
+     COMPUTE SigmV = sigma - vee.
+  END IF.

+  COMPUTE TXW=T(Xi)*W.
+  COMPUTE TTXW=T(TXW).
+  COMPUTE sumXWX = sumXWX + TXW*Xi.                /***sumXWX <- sumXWX + t(X)%*%W%*%X.
+  COMPUTE sumXWy = sumXWy + TXW*Yi.                /***sumXWy <- sumXWy + t(X)%*%W%*%Y.
+  COMPUTE sumXWJWX = sumXWJWX + TXW*J*TTXW.        /***sumXWJWX <- sumXWJWX + t(X)%*%W%*%J%*%W%*%X.
+  COMPUTE sumXWVWX = sumXWVWX + TXW*vee*TTXW.      /***sumXWVWX <- sumXWVWX + t(X)%*%W%*%vee%*%W%*%X.
+  COMPUTE sumXWsig = sumXWsig + TXW*SigmV*TTXW.    /***sumXW.sig.m.v.WX <- sumXW.sig.m.v.WX + t(X)%*%W%*%SigmV%*%W%*%X.
+  COMPUTE curRow=curRow+Ki.
END LOOP.                                                            /***}.
                                    
COMPUTE b = INV(sumXWX)*sumXWy.                                      /***b <- solve(sumXWX)%*%sumXWy.
COMPUTE X=DES   .                                                    /***X <- data.matrix(X.full[-c(1)]).
COMPUTE pred = X*b.                                                  /***data$pred <- X%*%b.
COMPUTE e = Y - pred.                                                /***data$e <- data$effect.size - data$pred.
COMPUTE W = MDIAG(WT).                                               /***W <- diag(data$weights).
COMPUTE sumW = CSUM(WT).                                             /***sumW <- sum(data$weights).
COMPUTE Qe = t(e)*W*e.                                               /***Qe <-t(data$e)%*%W%*%data$e.

!IF (!RHO !EQ !NULL) !THEN
COMPUTE Rho=0.0 .
!ELSE
COMPUTE Rho=!Rho.
!IFEND

COMPUTE denom = sumW - Msum(diag(INV(sumXWX)*sumXWJWX)).             /***denom <- sumW - sum(diag(solve(sumXWX)%*%sumXWJWX)).
COMPUTE term1 = (Qe - N + Msum(diag(INV(sumXWX)*sumXWVWX))) /denom.  /***term1 <- (Qe - N + sum(diag(solve(sumXWX)%*%sumXWVWX))) /denom.
COMPUTE term2 = (Msum(diag(INV(sumXWX)*sumXWsig))) / denom.          /***term2 <- (sum(diag(solve(sumXWX)%*%sumXW.sig.m.v.WX ))) / deno.
COMPUTE tausq1 = term1 + rho*term2.                                  /***tau.sq1 <- term1 + rho*term2.

** Do sensitivity analysis for TauSq here **.
COMPUTE rhoVal={0:10}/10.
COMPUTE rhoTest=MAKE(1,11,0).
LOOP Inc=1 TO 11.
COMPUTE rhoTest(1,Inc)=term1 + rhoVal(1,Inc) * term2.
END LOOP.
COMPUTE SensAn={rhoVal;rhoTest}.

COMPUTE tausq=tausq1.
DO IF (tausq1<0).
COMPUTE tausq1=0.                                                     /***tau.sq <- ifelse(tau.sq1<0,0,tau.sq1).
END IF.
COMPUTE rwts = 1/(K &*(MEANV + tausq1)).                              /***data$r.weights <- 1/(data$k*(data$mean.v + tau.sq).
COMPUTE sumXWXr = 0.                                                 /***sumXWX.r <- 0.
COMPUTE sumXWyr = 0.                                                 /***sumXWy.r<- 0.
COMPUTE curRow=1.
LOOP i=1 TO N.
+  COMPUTE Ki      = K(curRow,1).
+  COMPUTE endRow  = curRow+Ki-1.
+  DO IF (Ki > 1).
+     COMPUTE Wi    = MDIAG(rwts(curRow:endRow,1)).
+  ELSE IF (Ki = 1).
+     COMPUTE Wi    = rwts(curRow:endRow,1)).
+  END IF.
+  COMPUTE Xi      = DES(curRow:endRow,:).
+  COMPUTE Yi      = Y(curRow:endRow,1).
+  COMPUTE TXW     = T(Xi)*Wi.
+  COMPUTE sumXWXr = sumXWXr + TXW*Xi.
+  COMPUTE sumXWyr = sumXWyr + TXW*Yi.
+  COMPUTE curRow  = curRow+Ki.
END LOOP.

COMPUTE br         = INV(sumXWXr)*sumXWyr.                      /***b.r <- solve(sumXWX.r)%*%sumXWy.r .
COMPUTE X          = DES .                                      /***X <- data.matrix(X.full[-c(1)].
COMPUTE predr      = X*br.                                      /***data$pred.r <- X%*%b.r .
COMPUTE er         = Y - predr.                                 /***data$e.r <- cbind(data$effect.size) - data$pred.r.

COMPUTE sumXWWXr   = 0.                                         /***sumXWeeWX.r <- 0.
COMPUTE curRow     = 1.
LOOP i=1 TO N.                                                  /***for (i in (1:N)) {.

+  COMPUTE KI      = K(curRow,1).
+  COMPUTE endRow  = curRow+KI-1.

+  COMPUTE erI     = er(curRow:endRow,1).
+  COMPUTE sighatr = erI*T(erI).                                /***sigma.hat.r <- tab$e.r%*%t(tab$e.r).

+  DO IF (Ki > 1).
+     COMPUTE W    = MDIAG(rwts(curRow:endRow,1)).
+  ELSE IF (Ki = 1).
+     COMPUTE W    = rwts(curRow:endRow,1)).
+  END IF.
+  COMPUTE X       = DES(curRow:endRow,:).

+  COMPUTE TXW     = T(X)*W.
+  COMPUTE sumXWWXr= sumXWWXr + TXW*sighatr*T(TXW).               /***sumXWeeWX.r <- sumXWeeWX.r + t(X)%*%W%*%sigma.hat.r%*%W%*%X.
+  COMPUTE curRow  = curRow+KI.

END LOOP.     

!IFEND  **/END OF NON-HIERARCHICAL **/.

**********************************************************************************************************************.
**********************************************************************************************************************.
**********************************************************************************************************************.

** COMMON TO BOTH METHODS **.
COMPUTE INVXWXR    = INV(sumXWXr).
COMPUTE VRr        = INVXWXR*sumXWWXr*INVXWXR.                    /***VR.r <- solve(sumXWX.r)%*%sumXWeeWX.r%*%solve(sumXWX.r).
COMPUTE SE         = MAKE(1,p+1,0).                               /***SE <- c(rep(0,p+1)).
LOOP i=1 TO p+1.                                                  /***for (i in (1:(p+1))) {.
+  COMPUTE SE(i)   = sqrt(VRr(i,i))*sqrt(N/(N-(p+1))).            /***SE[i] <-sqrt(VR.r[i,i])*sqrt(N/(N-(p+1)).
END LOOP.            



COMPUTE AVGNEFS=NEFS/N.
COMPUTE DF=N-(P+1).
COMPUTE RLAB=Params.
COMPUTE TValue=br&/T(SE). 
 
COMPUTE PofT=TCDF(TValue,DF).

COMPUTE TCrVal=TCRIT(1,1).
COMPUTE LCI=Br - TCrVal &* T(SE).
COMPUTE UCI=Br + TCrVal &* T(SE).

LOOP i=1 to P+1.
+  DO IF PofT(i,1) >.5.
+    COMPUTE PofT(i,1) =1- PofT(i,1) .
+  END IF.
+  COMPUTE PofT(i,1) =2*PofT(i,1) .
END LOOP.

!IF (!PRINT !EQ DEF) !THEN
COMPUTE BetaSE={br,T(SE),TValue, PofT ,LCI, UCI}.
COMPUTE CLAB={"Coef","SE", "T", "Pr > |T|", "95% Confidence", "Interval"}.
print BetaSE  /format !FMT
              /title = "Parameter Estimates and Robust Standard Errors"
              /space 2 
              /rnames=RLAB 
              /cnames=CLAB.
!IFEND

!IF (!PRINT !EQ EXP) !THEN
COMPUTE EXPB=EXP(br).
COMPUTE EXPL=EXP(LCI).
COMPUTE EXPU=EXP(UCI).
COMPUTE BetaSE={EXPB,TValue, PofT ,EXPL, EXPU}.
COMPUTE CLAB={"Coef", "T", "Pr > |T|", "95% Confidence", "Interval"}.
print BetaSE  /format !FMT
              /title = "Exponential Transformed Parameter Estimates and Test Statistics"
              /space 2 
              /rnames=RLAB 
              /cnames=CLAB.
!IFEND

PRINT NEFS    / FORMAT 'F4.0' 
              / TITLE "N Level 1".

PRINT N       / FORMAT 'F4.0' 
              / TITLE "N Level 2".

PRINT AVGNEFS / FORMAT 'F8.2' 
              / TITLE "Average Level 1 N".

PRINT DF      / FORMAT = "F4.0"
              / TITLE "T-Test DF".

PRINT tausq1   / format !FMT
              / title "Tau-squared estimate" .

!IF (!HWEIGHT !EQ !NULL) !THEN
PRINT RHO     / FORMAT 'F8.2'
              / TITLE "Assumed Rho".
!IFEND 

!IF (!HWEIGHT !NE !NULL) !THEN
PRINT omegasq / format !FMT
              / title "Omega-squared estimate" .
!IFEND 

PRINT Qe / FORMAT 'F8.3' 
              / TITLE "Weighted Residual Sum of Squares Qe".

!IF (!HWEIGHT !EQ !NULL) !THEN
COMPUTE TSensAn=T(SensAn).
COMPUTE TLab={"Rho", "Tau Sq"}.
PRINT TSensAn / format !FMT
             /title "Sensitivity analysis Tau.sq" 
             /cnames=TLab.
!IFEND 

* Optionally write residuals to Active File *.
!IF (!RESID !EQ T) !THEN
* Free some memory by blasting DESIGN *.
COMPUTE DES=MAKE(1,1,0).
GET ALLVARS
  / VARIABLES ALL 
  / FILE * 
  / NAMES=ALLV 
  / MISSING=ACCEPT
  / SYSMIS=0.
COMPUTE SVNAMES={ALLV,"Pred","Resid"} .
SAVE {ALLVARS, Predr, Er}
    / NAMES=SVNAMES
    / OUTFILE *  .
!IFEND .

END MATRIX. 
* Place file back into original order * .
SORT CASES BY @ORD@.
* DO CLEANUP * .
MATCH FILES FILE *  
          / DROP @STID@ @K@ @S@ @ORD@ @NMISS@ @MEFF@ @NP@ @NN@ @TDF@ @T_CRIT@ @WEIGHT@ INTERCEP .
exe.
!IFEND 
RESTORE.
!ENDDEFINE.

