"""
a library of useful utility routines for SPSS
Copyright (C) 2005 by SPSS Inc.

"""
__author__ =  'spss'
__version__=  '1.0'


import spss
from spssaux import VariableDict
import random


class Sample:
    """Abstract sample class to be subclassed, e.g. SampleWithReplacement

    Every subclass should override the DrawSample method to provide a
    useful generator.  Pass a Sample object to a Job object as its
    sample parameter, and the Job will iterate under the generator's control.
    """

    def __init__(self, iterations=1):
        self.__iterations = iterations

    def DrawSample(self, submit=True):
        """Dummy generator, used by Job to iterate once.

        """
        for i in range(self.__iterations):
            yield None


class SampleWithReplacement(Sample):
    """Draw a sample with replacement from the active data file, by computing a weight variable.

    By default, the population size is taken to be the number of cases in the active datafile.
    For the convenience of those who are drawing booststrap samples, the sample size is the
    same as the population size, unless stated otherwise.  However, the algorithm will correctly
    undersample or oversample.  Since the commands to draw the sample are always the same,
    they are made available as the attribute SampleWeightCommands, as well as through the DrawSample
    generator function.
    
    Arguments to initialize:
            sampleSize:         defaults to populationSize
            populationSize:     defaults to number of cases
            iterations:         defaults to 1
            sampleWeightVar:    defaults to "SampleWeight"
            weight:             defaults to True; in case "WEIGHT BY SampleWeight" is not desired
            filter:             defaults to True; in case "FILTER BY SampleWeight" is not desired

    Data attributes:
            NumberToSample, Population, SampleWeightCommands, SampleWeightVariable
    """

    def __init__(self, iterations=1, sampleSize=-1, populationSize=-1, sampleWeightVar="SampleWeight", doWeight=True, doFilter=True):
        M = sampleSize
        N = populationSize

        # if population size is not given, use the current file size
        if N < 1:
            try:
               N = spss.GetCaseCount()
            except:
                print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

            # for the convenience of bootstrappers, if sample size is not given use population size
            if M < 1:
                M = N

            self.__iterations = iterations
                    
            self.NumberToSample = M
            self.Population = N
            self.SampleWeightCommands = []
            self.SampleWeightVariable = sampleWeightVar
            self.Weight = doWeight
            self.Filter = doFilter

            if M < 1 or N < 1:
                return

            wt  = []
            if doWeight: wt += ["weight off."]
            if doFilter: wt += ["filter off."]
            wt += ["do if $casenum=1."]
            wt += ["- compute #samplesize=%s." % M ]
            wt += ["- compute #populationsize=%s." % N ]
            wt += ["end if."]
            wt += ["do if (#samplesize>0 and #populationsize>0)."]
            wt += ["- compute %s=rv.binom(#samplesize, 1/#populationsize)." % sampleWeightVar]
            wt += ["- compute #samplesize=#samplesize-%s." % sampleWeightVar]
            wt += ["- compute #populationsize=#populationsize-1."]
            wt += ["else."]
            wt += ["- compute %s=0." % sampleWeightVar]
            wt += ["end if."]
            if doWeight: wt += ["weight by %s." % sampleWeightVar]
            if doFilter: wt += ["filter by %s." % sampleWeightVar]
            
            self.SampleWeightCommands = wt


    def DrawSample(self, submit=True):
        """Generator yields commands to compute sample weight in SPSS.

        For bootstrap samples, the commands will always be the same, and can be
        obtained through the data attribute SampleWeightCommands.
        """
        for i in range(self.__iterations):
            if submit:
                try:
                    spss.Submit(self.SampleWeightCommands)
                except:
                    print "Could not submit:\n%s" % self.SampleWeightCommands    # also error number, etc.
                    print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
            yield self.SampleWeightCommands

class SampleGroupsWithReplacement(SampleWithReplacement):

    def __init__(self, groups, iterations=1, sampleSize="GroupSize", actualGroupSize="GroupSize", sampleWeightVar="SampleWeight", doWeight=True, doFilter=True):
        """Sample within groups.  groups should be a variable in the file.

        groupsize can be a number or a variable.  If it is a variable, the first value in each group is used.
        The first iteration includes initialization specific to selection within groups:
          the data are sorted by group,
          non-missing values are selected,
          the number in each group is computed.
        """

        self.__iterations = iterations
        self.SampleSize = sampleSize
        self.GroupSize = actualGroupSize
        self.SampleWeightCommands = []
        self.SampleWeightVariable = sampleWeightVar
        self.Weight = doWeight
        self.Filter = doFilter

        groupsyntax  = ["sort cases by %s ." % groups]
        groupsyntax += ["select if ~missing(%s) ." % groups]
        groupsyntax += ["aggregate "]
        groupsyntax += ["  /outfile=* mode=addvariables"]
        groupsyntax += ["  /break=%s" % groups]
        groupsyntax += ["  /%s=N ." % actualGroupSize]
        self.GroupSyntax = groupsyntax

        wt  = []
        if doWeight: wt += ["weight off."]
        if doFilter: wt += ["filter off."]
        wt += ["do if ($casenum = 1) or (%s ~= lag(%s))." % (groups, groups)]
        wt += ["- compute #samplesize = %s." % sampleSize ]
        wt += ["- compute #populationsize = %s." % actualGroupSize ]
        wt += ["end if."]
        wt += ["do if (#samplesize > 0) and (#populationsize > 0)."]
        wt += ["- compute %s = rv.binom(#samplesize, 1/#populationsize)." % sampleWeightVar ]
        wt += ["- compute #samplesize = #samplesize - %s." % sampleWeightVar ]
        wt += ["- compute #populationsize = #populationsize - 1 ."]
        wt += ["else."]
        wt += ["- compute %s = 0." % sampleWeightVar]
        wt += ["end if."]
        if doWeight: wt += ["weight by %s." % sampleWeightVar ]
        if doFilter: wt += ["filter by %s." % sampleWeightVar ]
        
        self.SampleWeightCommands = wt


    def DrawSample(self, submit=True):
        """Generator yields commands to compute sample weight in SPSS.

        The most recently generated commands can be obtained through the
        data attribute SampleWeightCommands.
        The first iteration includes initialization specific to selection
        within groups:  the data are sorted by group, the number in each group is computed.
        """
        # when i == 0 run the initialization in GroupSyntax as well as the weight commands
        # when i == 1 switch back to just sample weight commands
        # thereafter leave it alone
        commands = self.GroupSyntax + self.SampleWeightCommands
        for i in range(self.__iterations):
            if i == 1:
                commands = self.SampleWeightCommands
            if submit:
                try:
                    spss.Submit(commands)
                except:
                    print "Could not submit:\n%s" % commands    # also error number, etc.
                    print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
            yield commands

	
class JackKnife(Sample):
    """Sample omitting one subject at a time; typically one case at a time

    Omit one case at a time, or omit an entire group.  If weight is false, Filter is used.
    If there are G groups, weights may be adjusted to G/(G-1).
    """

    def __init__(self, sampleWeightVar="SampleWeight", doWeight=False, doFilter=True, subjectIdentifier="$casenum", numberOfSubjects=-1, subjects=[], indicatorCoding=True):
        self.SampleWeightVariable = sampleWeightVar
        self.Weight = doWeight
        self.Filter = doFilter
        self.SampleWeightCommands = []
        self.SubjectIdentifier = subjectIdentifier
        self.Subjects = subjects
        self.NumberOfSubjects = numberOfSubjects
        self.useIndicatorCoding = indicatorCoding
        #print "1-Num : %s" % self.NumberOfSubjects

        lsub = len(subjects)
        if lsub > 0:
            self.NumberOfSubjects = lsub
        #print "2-Num : %s" % self.NumberOfSubjects
        
        if self.NumberOfSubjects < 1:
            try:
                self.NumberOfSubjects = spss.GetCaseCount()
            except:
                self.NumberOfSubjects = 0
        #print "3-Num : %s" % self.NumberOfSubjects
        
        if self.NumberOfSubjects > 0 and self.Subjects == []:
            self.Subjects = xrange(1, self.NumberOfSubjects + 1)


    def DrawSample(self, submit=True):
        """Generator yields commands to compute sample weight in SPSS.

        The most recent commands can be obtained through the data attribute SampleWeightCommands.
        """

        if self.Subjects == [] or self.NumberOfSubjects < 1:
            return
        
        w = float(self.NumberOfSubjects)
        if not self.useIndicatorCoding: w = w/(w-1)
        
        wt0  = []
        if self.Weight: wt0 += ["weight off."]
        if self.Filter: wt0 += ["filter off."]

        wt1 = ["compute %s=(%s~=" % (self.SampleWeightVariable, self.SubjectIdentifier)]
        wt1 += ["%s)"]    # defer substitution until the loop below
        if self.Weight:
            wt1 += ["*%s ." % w ]
        else:
            wt1 += [" ."]
        wt1 = "".join(wt1)
        
        wt2 = []
        if self.Weight: wt2 += ["weight by %s." % self.SampleWeightVariable]
        if self.Filter: wt2 += ["filter by %s." % self.SampleWeightVariable]

        for i in self.Subjects:
            wt = []
            wt += wt0
            wt += [wt1 % repr(i)]   # deferred substitution happens here -- repr adds quotes for string types
            wt += wt2
            self.SampleWeightCommands = wt
            if submit:
                spss.Submit(wt)
            yield wt

	
# ToDo: should be a classmethod of Bootstrap...
def accelerationMultiplier(n):
    """Computes an adjustment needed to convert between skewness and acceleration.

    Multiply the skewness times the result to obtain acceleration.
    The argument should be an integer > 2"""
    from math import sqrt
    return -sqrt(n-1.0)*(n-2.0)/(6.0*n)


# ToDo: should be a classmethod or something...
def selectCompleteCases(variables=[], completeDataset="complete__"):
    """Convenience method to produce syntax to select complete cases.

    Copies the active dataset, selects only cases which are non-missing
    for all variables specified, and activates the copy.
    """
    select  = ["dataset copy %s ." % completeDataset ]
    select += ["dataset activate %s ."% completeDataset ]
    select += ["select if (nmiss(%s) = 0) ." % ", ".join(variables)]
    return select

def listJoin(obj, separator=" "):
    """Convenience method joins list elements using separator

    Simplifies functions which allow either string or list as arguments"""
    # first make everything into a list
    if isinstance(obj, str):
        lobj = [obj]
    elif isinstance(obj, (list, tuple)):
        lobj = obj
    else:
        lobj = list(obj)
    # but always return a string, using separator to join the elements
    return separator.join(lobj)


class OMS:
    """Encapsulates OMS syntax, manages XML Workspace, and extracts results from output.

    Results are returned in a dictionary of lists; a dataset object may be created also.
    
    Basic usage:  supply commands and subtypes, then request values, supplying a name for each.
    Use Submit() to activate the request(s), and end() to create an XML workspace.
    At this point, use getValues() to obtain a dictionary indexed by the request names.

    For example, in a regression with variables x1 and x2:
    
    oms = OMS("Regression", "Coefficients")
    oms.requestValues("constant", rowCategory="(Constant)", colCategory="B")
    oms.requestValues("b1", rowCategory="x1", colCategory="B")
    oms.requestValues("b2", rowCategory="x2", colCategory="B")

    # If used in conjunction with a Job object, the following is handled automatically.
    
    oms.Submit()
    spss.Submit("REGRESSION /DEPENDENT y /METHOD=ENTER x1 x2 ."
    oms.end()
    values = oms.getValues()
    print values
    #create a Dataset object and make it active
    regCoeff = oms.createDataset("Reg_Coeff", activate=True)

    The values dictionary will have keys ["constant", "b1", "b2"]; each item in the dictionary
    will be a list with one element.  If more than one Regression command is submitted, the
    lists will contain an element for each regression.

    There is a rudimentary state machine built in which tries to track active requests
    versus those which have completed, and therefore have an XML Workspace available.    
    """

    activeRequests = {}

    def __init__(self, commands=[], subtypes=[], tag="", select="TABLES",
                 format="OXML", xmlworkspace=True, datasetname="",
                 savoutfile="", columnDimensions=[],
                 viewer=False, suppressAll=True):
        if tag == "":
            self.tag = "S%s" % random.uniform(0,1)
        else:
            self.tag = tag
        self.select = select
        self.__commands = commands
        self.subtypes = subtypes
        self.format = format
        self.xmlworkspace = xmlworkspace
        self.viewer = viewer
        #self.xpathselector = ""
        self.keys = []     # typically this would be a list of variable names
        self.xpath = {}    # xpath used to request values from xmlworkspace
        self.values = {}   # store a value for each key after retrieving it from the workspace
        self.dataset = datasetname
        self.suppressall = True
        self.outfile = None
        self.__complete = False
##        self.__active = False
        
        o = ["OMS"]
        if select != "":
            o += [" /SELECT %s" % select ]
        if commands or subtypes:
            ifclause = [" /IF"]
            if commands:
                if not isinstance(commands, list): commands = [commands]
                ifclause += ["COMMANDS=[%s]" % " ".join(map(repr, commands)) ]
            if subtypes:
                if not isinstance(subtypes, list): subtypes = [subtypes]
                ifclause += ["SUBTYPES=[%s]" % " ".join(map(repr, subtypes)) ]
            o += [" ".join(ifclause)]

##          # For some reason the above works, but what is below does not
##            if commands:
##                if not isinstance(commands, list):
##                    self.__commands = [commands]
##                else:
##                    self.__commands = commands
##                ifclause += ["COMMANDS=[%s]" % " ".join(map(repr, self.__commands)) ]
##            if subtypes:
##                if not isinstance(subtypes, list):
##                    self.subtypes = [subtypes]
##                else:
##                    self.subtypes = subtypes
##                ifclause += ["SUBTYPES=[%s]" % " ".join(map(repr, self.subtypes)) ]
##          o += [" ".join(ifclause)]
        dest = [" /DESTINATION FORMAT=%s " % format]
        if xmlworkspace:
            dest += ["XMLWORKSPACE='%s'" % self.tag] #xmlworkspace]
        if savoutfile != "":
            dest += ["OUTFILE=%s", savoutfile]
        dest += ["VIEWER=%s" % (viewer and "YES" or "NO")]
        dest = " ".join(dest)
        o += [dest]
        #o += [" /DESTINATION FORMAT=%s XMLWORKSPACE='%s' VIEWER=%s" % (format, xmlworkspace, viewerYesNo)]
        if columnDimensions != []:
            o += [" /COLUMNS DIMNAMES=[%s]" % " ".join(map(repr, columnDimensions)) ]
        # ToDo: outputset
        o += [" /TAG='%s' ." % self.tag]

        self.omsend = ["OMSEND TAG='%s'." % self.tag]
        if suppressAll:
            self.oms = ["OMS /DESTINATION VIEWER=NO /TAG='%s_suppressall'." % self.tag]
            self.oms += o
            self.omsend += ["OMSEND TAG='%s_suppressall'." % self.tag]
        else:
            self.oms = o

    def commands(self):
        """ToDo:  move most of the command construction here, instead of putting it in __init__"""
##        viewerYesNo = self.viewer and "YES" or "NO"
##        subst = {"viewerYesNo":viewerYesNo}
##        self.dest
##        oms = self.oms % subst  ### needs to be specific about dest
        return self.oms

    def end(self):
        """OMS requests remain active until ended.  Be sure that end() is called!

        After end(), the XML Workspace will be available for querying with XPath.
        """
        self.__complete = True
        try:
            if OMS.activeRequests.pop(self.tag, None) != None:
                spss.Submit(self.omsend)
            else:
                print "OMS request %s is not active." % self.tag
                print "\n".join(self.oms)
                # should we try to inactivate it anyway?
        except:
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

            
##        if self.__active:
##            self.__active = False
##            spss.Submit(self.omsend)
##            self.__complete = True

    def getKeys(self):
        """Each OMS request has a key, typically used as a variable name.

        It is the key used to retrieve the result after getValues() is called.
        If a Dataset object is used to save the results, it will become the
        variable name.
        """
        return self.keys

    def suppressAll(self, suppress=True):
        """Turns off the viewer for all output

        May become an object in its own right someday.
        """
        suppressTag = "%s_suppressall"
        suppressAll = "OMS /DESTINATION VIEWER=NO /TAG='%s'." % suppressTag

        if suppress:
            try:
                spss.Submit(suppressAll)
            except:
                print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
        return suppressTag

    def Submit(self):
        """OMS requests are not active until Submit() is called.

        This will be handled automatically when used in conjunction with a Job object.
        """
        #print "OMS request %s submitting commands:\n%s\n" % (self.tag, "\n".join(self.oms))
        try:
            spss.Submit(self.commands())
            # consider using a weak reference here
            OMS.activeRequests[self.tag] = self
            ##self.__active = True
        except:
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
                
   
    def __XPathSelector(self, category, attrib):
        """Create XPath qualifier fragment from a category and value
        If category is a list, a sequence of nested predicates is constructed
        """
        if category:
            c = category
            if not isinstance(c, (list, tuple)):
                c = [c]
            res = []
            for i in c:
                res += "//category[@%s='%s']" % (attrib, i)
            res = "".join(res)
        else:
            res =  "//category"
        return res

    def requestValues(self, key, tableSubtype="", rowCategory=None, rowAttrib="text",
        colCategory=None, colAttrib="text", cellAttrib="number"):
        """Construct an xpath expression, to be used later by getValues.

        key is typically a variable name, used to retrieve the value later.
        tableSubtype is the subtype of the table from which to retrieve data.  It is the
        OMS table type identifier.
        rowCategory is the optional row category value to fetch.  If omitted, all rows
        meeting the other criteria are returned.
        rowAttrib is the attrib to use with rowCategory.  It defaults to text but may
        be any type in the table row labels such as label or varName
        colCategory is the optional column category to fetch.  If omitted, all columns
        meeting the other criteria are returned.
        colAttrib is the attribute type for colCategory.  It defaults to text but may
        be any type in the column labels such as label or varName
        cellAttrib is the attribute type to return.  It defaults to number, but it may
        usefully be specified as all ("*"), number, date, or text.
        """
        self.keys += [key]
        if tableSubtype == "":
            subtype = self.subtypes
        else:
            subtype = tableSubtype
        xp  = ["//pivotTable[@subType=%s]" % repr(subtype) ]
        xp += ["//dimension[@axis='row']%s" % self.__XPathSelector(rowCategory,rowAttrib)]
        xp += ["//dimension[@axis='column']%s" % self.__XPathSelector(colCategory,colAttrib)]
        xp += ["/cell/@%s" % cellAttrib]
        xp = "".join(xp)
        self.xpath[key] = xp
        return xp

    def requestRows(self, rows=None, keys=None, tableSubtype="", rowAttrib="text",
        colCategory=None, colAttrib="text", cellAttrib="number"):
        """Convenience method to repeatedly call requestValues for a list or dictionary.

        If a list is given for rows, the row attribute is also used as the key.  This will not always
        be an acceptable key value.  If a dictionary is given, keys may also be supplied in a list.
        This allows the order in which the keys are enumerated to be controlled; also, subsets of the
        dictionary elements may be selected.

        ToDo: requestColumns()
        """
        if isinstance(rows, (list, tuple)):
            for key in rows:
                self.requestValues(key, tableSubtype, key, rowAttrib, colCategory, colAttrib, cellAttrib)
        elif isinstance(rows, dict):
            if isinstance(keys, (list, tuple)) and rows != {}:
                for key in keys:
                    self.requestValues(key, tableSubtype, rows[key], rowAttrib, colCategory, colAttrib, cellAttrib)
            else:
                for key, row in rows.iteritems():
                    self.requestValues(key, tableSubtype, row, rowAttrib, colCategory, colAttrib, cellAttrib)

    def requestGroup(self, key_group, key_group_value, group="Percentiles", cellAttrib="number"):
        """Customized version of requestValues when there are group labels, specifically Percentiles

        ToDo: this has not been tested for any Table but "Descriptives", "Descriptive Statistics"
        nor for any group but "Percentiles".  It does what is needed for Bootstrap bcaPercentileI

        ToDo: find the xpath to get percentiles from Frequencies
        """
        # xp = "//pivotTable[@subType='Statistics']//dimension[@axis='row']//group[@text='Percentiles']/category/@number"
        # xptile = "//pivotTable[@subType='Statistics']//dimension[@axis='row']//group[@text='Percentiles']/category//cell/@number"
        self.keys += [key_group]
        xpath =  ["//pivotTable[@subType='%s']" % self.subtypes ]
        xpath += ["//dimension[@axis='row']"]
        xpath += ["//group[@text='%s']" % group]
        xpath += ["/category"]
        xp = xpath[:]   # slice copy
        xpath += ["/@%s" % cellAttrib]
        xpath = "".join(xpath)
        self.xpath[key_group] = xpath
        self.keys += [key_group_value]
        xpath = xp + ["//cell/@%s" % cellAttrib ]
        xpath = "".join(xpath)
        self.xpath[key_group_value] = xpath
        return xpath
    
    def getValues(self):
        """Add values to a dictionary by evaluating xpath"""
        self.values = {}

        if not self.__complete:
            print "OMS request with tag %s is not active.  Call OMS.Submit() and OMS.end() before getValues." % self.tag
            # ToDo: really should raise an exception here
        for key, xp in self.xpath.iteritems():
            try:
                evalxp = spss.EvaluateXPath(self.tag, "/outputTree", xp)
                self.values[key] = evalxp
            except:
                print "xpath could not be evaluated for handle %s:\n%s" % (self.tag, xp) # also error number, etc.
                print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
        # return values
        return self.values

    def deleteXMLWorkspace(self):
        if self.tag in OMS.activeRequests:
            self.end()
        try:
            spss.DeleteXPathHandle(self.tag)
        except:
            print "xpath could not be deleted for handle %s" % self.tag
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

    def createDataset(self, dataset, variableLabels={}, submit=True, activate=False, deleteWorkspace=True):
        """Send the OMS values to a workspace.  Not valid until after end() has been called.

        The dataset may be retrieved using "DATASET ACTIVATE" and the dataset name supplied,
        or with self.outfile.activate()
        """
        if not self.__complete:
            print "createDataset cannot be called before OMS.Submit() and OMS.end()."
            # ToDo: really should raise an exception
        out = Dataset(dataset, self.keys, varLabels=variableLabels)
        values = self.getValues()
        #### ToDo: decide when to use self.values
        out.addDataFromDictionary(values)
        if submit: out.Submit()
        if activate: out.activate()
        self.outfile = out
        if deleteWorkspace:
            self.deleteXMLWorkspace()
        return out

	
class Dataset:
    """Use with OMS_XML to send output to a dataset

    Constructs a file using DATA LIST.  All data are entered as strings,
    though only numeric variables are currently supported.
    THIS CLASS MAY SOON BE REPLACED by one which uses the I/O dll to
    write directly to a data file.

    A rudimentary state machine supports adding rows one at a time with addRow().
    At this time, the predominant mode of operation uses addDataFromDictionary(),
    which doesn't demand state information.

    ToDo: add support for string variables, value labels
    ToDo: classmethod for closing datasets automatically
    """
    #states = ["Initialized", "Created", "Activated", "Closed"]
    Initialized = 0
    Created = 1
    Activated = 2
    Closed = 3
    
    # maintain the most recently activated dataset, if known
    activeDataset = None
    datasets = {}
    
    def __init__(self, dataset, varNames, varLabels = {}):
        # maintain a dictionary of dataset names as they are created
        Dataset.datasets[dataset] = self
        self.data = []
        self.variables = varNames
        self.variableLabels = varLabels
        
        # construct data list command
        if isinstance(varNames, (list, tuple)):
            self.numberOfVariables = len(varNames)
            cmd  = ["data list free"]
            cmd += ["   /%s" % " ".join(varNames)]
            cmd += ["."]
            cmd += ["begin data"]
            self.begindata = cmd

            cmd = ["end data."]
            # create variable labels
            if len(varLabels) > 0:
                cmd += ["variable labels"]
                cmd += [" %s '%s'" % (name, label) for name, label in varLabels.iteritems()]
                cmd += ["."]
            self.enddata = cmd
        else:
            self.numberOfVariables = 0
            self.begindata = ""
            self.enddata = ""
        self.dataset = dataset
        self.__dataset = ["dataset name %s ." % dataset]
        self.__state = Dataset.Initialized
        self.__exists = False
        self.__closed = False

    def addRow(self, datarow):
        """Add another row to the DATA LIST"""
        if isinstance(datarow, (list, tuple)):
            self.data += [" ".join(map(str, datarow))]
        else:
            self.data += [str(datarow)] # could check that self.numberOfVariables == 1

    def addDataFromDictionary(self, dataDictionary={}):
        """Dictionary should contain a list of values for each variable.

        There must be a key for each variable in varNames.
        ToDo: enforce this restriction and/or raise exceptions.
        """
        if isinstance(dataDictionary, dict):
            datacols = [dataDictionary[k] for k in self.variables]
            datarows = zip(*datacols)
            self.data += [" ".join(map(str, row)) for row in datarows]

    def commands(self):
        """Returns the commands which Submit() will execute."""
        cmds = self.begindata + self.data + self.enddata + self.__dataset
        return cmds
    
    def Submit(self):
        """Tell SPSS to submit the commands, creating the dataset."""
        try:
            spss.Submit(self.commands())
            self.__state = Dataset.Created
            self.__exists = True
            self.__closed = False
        except:
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

    def activate(self):
        """Make this dataset the active one in SPSS."""
        if self.__closed or not self.__exists: return
        try:
            spss.Submit("dataset activate %s ." % self.dataset)
            activeDataset = self
            self.__state = Dataset.Activated
            Dataset.datasets[self.dataset] = self
        except:
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
        
    def close(self):
        """Close this dataset in SPSS."""
        if not self.__exists: return
        if self.__state in (Dataset.Created, Dataset.Activated):
            try:
                spss.Submit("dataset close %s ." % self.dataset)
                if Dataset.activeDataset is self:  Dataset.activeDataset = None
                self.__state = Dataset.Closed
                Dataset.datasets.pop(self.dataset, None)
                self.__exists = False
                self.__closed = True
            except:
                print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

    def save(self, filename):
        """Save the dataset to a file."""
        if self.__closed or not self.__exists: return
        if self.__state in (Dataset.Initialized, Dataset.Closed): return
        self.activate()
        try:
            spss.submit("save outfile='%s'." % filename)
        except:
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

    def summaryTable(self, varnames, title="Summary Table", stats="MEAN, MEDIAN, STDDEV", submit=True):
        """Display the dataset values using the Summarize command instead of CTables"""
        if Dataset.activeDataset is not self:
            self.activate()
        varlist = listJoin(varnames)
        statlist = listJoin(stats)
        
        summarize  = ["summarize"]
        summarize += ["   /tables=%s" % varlist]
        summarize += ["   /format=nolist nototal"]
        summarize += ["   /title='%s'" % title]
        summarize += ["   /cells=%s ." % statlist]
        if submit: spss.Submit(summarize)
        return summarize

    def summaryCTable(self, varnames, title="Summary Table", stats="MEAN, MEDIAN, STDDEV", submit=True):
        """Display the dataset values in a Summary table using CTables"""
        # activate the dataset if not already active...
        if Dataset.activeDataset is not self:
            self.activate()
##        if isinstance(varnames, str):
##            variables = varnames
##            tablevarlist = varnames
##        elif isinstance(varnames, (list, tuple):
##            variables = " ".join(varnames)
##            tablevarlist = "+".join(varnames)
        varlist = listJoin(varnames)
        tablevarlist = listJoin(varnames, "+")
        statlist = listJoin(stats)

        lines = ["CTABLES"]
        lines+= ["  /VLABELS VARIABLES=%s DISPLAY=DEFAULT" % varlist]
        lines+= ["  /TABLE (%s) [S][%s]" % (tablevarlist, statlist)]
        lines+= ["  /TITLES  TITLE= '%s'." % title]
        if submit: spss.Submit(lines)
        return lines

    def summaryTableBy(self, row, column, varnames, title="Summary Table", stats="MEAN, MEDIAN, STDDEV", submit=True):
        """Display the dataset values using the Summarize command instead of CTables"""
        if Dataset.activeDataset is not self:
            self.activate()
        rowlist = listJoin(row)
        collist = listJoin(column)
        varlist = listJoin([rowlist, collist, listJoin(varnames)])
        tablevarlist = listJoin(varnames, "+")
        statlist = listJoin(stats)
        summarize  = ["summarize"]
        summarize += ["   /tables=%s" % listJoin([varlist, rowlist, collist], " by ")]
        summarize += ["   /format=nolist nototal"]
        summarize += ["   /title='%s'" % title]
        summarize += ["   /cells=%s ." % statlist]
        if submit: spss.Submit(summarize)
        return summarize

    def summaryCTableBy(self, row, column, varnames, title="Summary Table", stats="MEAN, MEDIAN, STDDEV", submit=True):
        """Display the dataset values in a Summary table of row variable(s) by column(s) """
        # activate the dataset if not already active...
        if Dataset.activeDataset is not self:
            self.activate()
##        if isinstance(varnames, str):
##            variables = varnames
##            tablevarlist = varnames
##        elif isinstance(varnames, (list, tuple):
##            variables = " ".join(varnames)
##            tablevarlist = "+".join(varnames)
##        varlist = " ".join([row]+[column]+varnames)
##        tablevarlist = "+".join(varnames)
        rowlist = listJoin(row)
        collist = listJoin(column)
        varlist = listJoin([rowlist, collist, listJoin(varnames)])
        tablevarlist = listJoin(varnames, "+")
        statlist = listJoin(stats)
        lines = ["CTABLES"]
        lines+= ["  /VLABELS VARIABLES=%s DISPLAY=DEFAULT" % varlist]
        lines+= ["  /TABLE %s BY %s > %s [%s]" % (rowlist, collist, tablevarlist, statlist)]
        lines+= ["  /CATEGORIES VARIABLES=%s %s ORDER=A KEY=VALUE EMPTY=INCLUDE" % (rowlist, collist)]
        lines+= ["  /TITLES  TITLE= '%s'." % title]
        if submit: spss.Submit(lines)
        return lines

class MakeDataset(Dataset):
    """Creates datasets using arbitrary Compute statements.

    Basic usage:

    data = MakeDataset("phony_data", cases=1000)
    data.addCompute("x", "uniform(1)")
    data.addCompute("y", "rv.normal(0,1)")
    data.Submit()

    Useful for testing and simulation.
    """

    def __init__(self, dataset="created_data__", cases=100):
        Dataset.datasets[dataset] = self
        self.dataset = dataset
        self.cases = cases
        self.compute = []

        self.data = []
        self.variables = []
        self.variableLabels = {}
        self.dataset = dataset
        self.__dataset = ["dataset name %s ." % dataset]
        self.__state = Dataset.Initialized
        self.__exists = False
        self.__closed = False
        
    def addCompute(self, var, expression, label=""):
        """ToDo: replace this with Transform objects"""
        self.variables += [var]
        self.compute += ["compute %s = %s ." % (var, expression)]
        if label != "": self.variableLabels[var] = label
        
    def Submit(self, submit=True, activate=True):
        mkdata  = ["new file."]
        mkdata += ["input program."]
        mkdata += ["loop #i = 1 to %s." % self.cases]
        mkdata += self.compute
        mkdata += ["+ end case."]
        mkdata += ["end loop."]
        mkdata += ["end file."]
        mkdata += ["end input program."]
        mkdata += ["execute."]

        if self.variableLabels != {}:
            mkdata += ["variable labels"]
            mkdata += ["   %s '%s'" % (var, label) for var, label in self.variableLabels.iteritems()]
            mkdata += ["."]
        mkdata += ["execute ."]
        mkdata += ["dataset name %s ." % self.dataset]
        if submit:
            try:
                spss.Submit(mkdata)
            except:
                print "Failed to create dataset ", self.dataset
                print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())
            if activate: self.activate()
        return mkdata

    def commands(self):
        return self.Submit(False, False)
    
    def activate(self):
        #if self.__closed or not self.__exists: return
        try:
            spss.Submit("dataset activate %s ." % self.dataset)
            activeDataset = self
            self.__state = Dataset.Activated
            Dataset.datasets[self.dataset] = self
        except:
            print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

    def close(self):
        #if not self.__exists: return
        if self.__state in (Dataset.Created, Dataset.Activated):
            try:
                spss.Submit("dataset close %s ." % self.dataset)
                if Dataset.activeDataset is self:  Dataset.activeDataset = None
                self.__state = Dataset.Closed
                Dataset.datasets.pop(self.dataset, None)
                self.__exists = False
                self.__closed = True
            except:
                print "Error level %s: %s" % (spss.GetLastErrorLevel(), spss.GetLastErrorMessage())

class Job:
    """Organizes syntax, OMS, iterators, XMLWorkspaces, and datasets into a single "job".

    Syntax is assumed to be held in a string or a list of strings.
    OMS objects are typically supplied to manage output.
    Sample objects are used to iterate.
    Pre and Post actions may be added to allow program action to be taken before and after
    running the syntax, for each iteration.

    Basic usage:

        syntax = "REGRESSION /DEPENDENT y /METHOD=ENTER x1 x2 ."
        
        oms = OMS("Regression", "Coefficients")
        oms.requestValues("constant", rowCategory="(Constant)", colCategory="B")
        oms.requestValues("b1", rowCategory="x1", colCategory="B")
        oms.requestValues("b2", rowCategory="x2", colCategory="B")

        job = Job(syntax, oms)
        values = job.Submit()
        # do something with the values
        print values
        
        #create a Dataset object and make it active
        regCoeff = job.createDataset("Reg_Coeff", activate=True)


    For a basic Bootstrap with 100 iterations, simply use instead:

        job = Job(syntax, oms, SampleWithReplacement(100))

    but consider using the Bootstrap object, which will also compute
    bias-corrected confidence intervals.    
    """

    def __init__(self, syntax=[], oms=[], sample=None):
        if isinstance(syntax, (list, tuple)):
            self.syntax = syntax
        elif isinstance(syntax, str):
            self.syntax = [syntax]
        else:
            self.syntax = []
        if isinstance(oms, (list, tuple)):  # should check that elements are OMS instances !
            self.__oms = oms
        elif isinstance(oms, OMS):
            self.__oms = [oms]
        else:
            self.__oms = []        
        self.__iterator = sample    # check that it has attribute DrawSample...
        self.__delegates_pre = []
        self.__delegates_post = []
        self.__values = None

                
    def addOMS(self, oms):
        if isinstance(oms, OMS):
            self.__oms.append(oms)

    def getOMS(self):
        return self.__oms

    def addPreAction(delegate):
        """Any object supplied as delegate must be able to respond to Submit()

        Delegates designated PreAction are called before the syntax is submitted.
        """
        if getattr(delegate, "Submit", None) is not None:
            self.__delegates_pre.append(delegate)

    def addPostAction(delegate):
        """Any object supplied as delegate must be able to respond to Submit().

        Delegates designated PostAction are called after the syntax is submitted,
        """
        if getattr(delegate, "Submit", None) is not None:
            self.__delegates_post.append(delegate)

    def Submit(self):
        """Submits oms requests, pre-actions, syntax, post-actions, ends oms requests.

        Note that the workspaces will not remain active.
        """
        self.__values = None
        for oms in self.__oms:
            oms.Submit()
        if self.__iterator is None:
            self.__iterator = Sample()
        for sample in self.__iterator.DrawSample():
            for delegate in self.__delegates_pre:
                delegate.Submit()
            spss.Submit(self.syntax)
            for delegate in self.__delegates_post:
                delegate.Submit()
        # there's really no reason to end the requests in reverse order,
        # but it seems like the right thing to do anyway
        for oms in reversed(self.__oms):
            oms.end()
        self.__values = self.getValues()
        for oms in self.__oms:
            oms.deleteXMLWorkspace()
        return self.__values

    def getValues(self):
        if self.__values == None:
            values = {}
            for oms in self.__oms:
                values.update(oms.getValues())
            self.__values = values
        else:
            values = self.__values
        return values
       
    def __getitem__(self, var):
        """ToDo: Eventually it will be possible to use the Job object itself as a dictionary,
        after Submit() is called,
        """ 
        values = None
        try:
            values = self.__values[var]
        except:
            pass
        return values
            
    def createDataset(self, dataset, variableLabels={}, activate=False):
        """Send the OMS values to a dataset. Variable Labels may be supplied.

        The dataset may be retrieved using "DATASET ACTIVATE" and the dataset name supplied,
        or with self.outfile.activate().  Variable labels for variables in the
        dictionary but not in the oms keys will be ignored.
        """
        variables = []
        for oms in self.__oms:
            variables += oms.getKeys()
        
        out = Dataset(dataset, variables, varLabels=variableLabels)
        values = self.getValues()
        out.addDataFromDictionary(values)
        out.Submit()
        if activate: out.activate()
        return out
    

class Descriptives(Job):
    """Send Descriptive Statistics to an XML workspace and/or Dataset.

    allowed statistics are lists of strings
    MEAN SUM STDDEV VARIANCE RANGE MIN MAX SEMEAN KURTOSIS SKEWNESS
    but they appear in the table as Mean, Sum, Std., Variance...
    also if skewness or kurtosis is requested,
    for colCategory need to use ["Skewness", "Statistic"]
    Note that, unlike syntax, SEKurtosis and SESkewness can be used to
    request the standard errors of Skewness and Kurtosis.
    """
    statsDictionary = {
        "Mean":["Mean"],
        "Sum":["Sum"],
        "Stddev":["Std. Deviation"],
        "Variance":["Variance"],
        "Range":["Range"],
        "Min":["Minimum"],
        "Max":["Maximum"],
        "SEMean":["Mean"],
        "Kurtosis":["Kurtosis"],
        "Skewness":["Skewness"],
        "SEKurtosis":["Kurtosis"],
        "SESkewness":["Skewness"]}
    requiresNesting = ("SEMean","Kurtosis","Skewness","SEKurtosis","SESkewness")
    hasStdError = ("SEMean","SEKurtosis","SESkewness")

    def __init__(self, variables, statistics=["Mean"], variableLabels={}, sample=None):
        stats = [stat for stat in statistics if stat in Descriptives.statsDictionary]
        stats0 = [stat for stat in stats if stat not in ("SESkewness","SEKurtosis")]
        syntax  = ["Descriptives"]
        syntax += ["    variables = %s " % " ".join(variables)]
        syntax += ["   /Statistics %s ." % " ".join(stats0)]

        nestedTable = False
        for stat in statistics:
            if stat in Descriptives.requiresNesting:
                nestedTable = True
        colCat = {}
        # nested tables require a list of categories;
        # here the second category is either "Statistic" or "Std. Error"
        if nestedTable:
            for stat in stats:
                if stat in Descriptives.hasStdError:
                    colCat[stat] = Descriptives.statsDictionary[stat]+["Std. Error"]
                elif stat == "Stddev":
                    colCat[stat] = ["Std.","Statistic"]
                else:
                    colCat[stat] = Descriptives.statsDictionary[stat]+["Statistic"]
        else:
            # for non-nested tables just look up the label in the dictionary
            for stat in stats:
                colCat[stat]= Descriptives.statsDictionary[stat]
        
        oms = OMS(commands="Descriptives", subtypes="Descriptive Statistics")
        for stat in statistics:
            if variableLabels == {}:
                oms.requestRows(variables, colCategory=colCat[stat])
            else:
                oms.requestRows(variableLabels, variables, colCategory=colCat[stat])
        #print "Descriptives Job OMS: ", oms.commands()
        #print syntax
        Job.__init__(self, syntax, oms, sample)


class Bootstrap(Job):
    """Subclass of Job which knows how to compute bias-corrected accelerated confidence intervals.

    Basic usage:
        oms = OMS(commands="Regression", subtypes="Coefficients")
        oms.requestValues("constant", rowCategory="(Constant)", colCategory="B")
        oms.requestRows(vlabels, indvars, colCategory="B")

        job = Bootstrap(syntax, oms, iterations=100)
        bootrep = job.Submit()

        # summary table will be much nicer if CTables can be used
        job.bcaSummary(useCTables=True)
        job.bcaSummary(useCTables=False)
    """
    def __init__(self, syntax, oms, iterations=1, completeDataset="complete__", sample=None, BCaConfidenceIntervals=True):
        self.iterations = iterations
        self.completeDataset = completeDataset
        if sample is None:
            self.sample = SampleWithReplacement(iterations)
        self.values = {}
        self.bcaCI = BCaConfidenceIntervals
        self.bcaDataset = None
        Job.__init__(self, syntax, oms, sample)

    def Submit(self):
        spss.Submit(["dataset activate %s ." % self.completeDataset])
        job0 = Job(self.syntax, self.getOMS())
        theta = job0.Submit()
        self.values = theta

        # would like to be able to switch oms.viewer = False here
        # ToDo: need to update OMS object to support this
        job1 = Job(self.syntax, self.getOMS(), self.sample)
        job1.Submit()
        bootrep = job1.createDataset("bootstrap_replicates__", activate=True)
        if self.bcaCI:
            self.bcaDataset = self.bcaConfidenceIntervals(bootrep)
        return bootrep

    def bcaConfidenceIntervals(self, dataset, bc_dataset="z0__"):
        """Computes Bias-Corrected and Accelerated Confidence Intervals.

        See Efron and Tibshirani, An Introduction to the Bootstrap, Chapter 14
        The code here is ugly and subject to change, but may be useful to
        illustrate how the various objects can be used.
        """
        #
        # COMPUTE THE BIAS-CORRECTION, z0
        #
        # we need to compute the proportion less than the original value
        # first compute some indicator variables
        # remember long ago we saved the original coefficient statistics as theta
        # theta is a dictionary of lists, we happen to know there is only one element
        # in each list, so we ask for theta[var][0]
        theta = self.values
        z0syntax  = []
        z0vars = []
        z0rows = {}
        # the following dictionary will be useful later
        z0key = {}
        # this could and probably will be replaced by transformations from spssaux
        for var in dataset.variables:
            z0var = "%s_z0" % var
            z0key[var] = z0var
            z0rows[z0var] = z0var
            z0vars += [z0var]
            # use dictionary substitution into the string below
            subst = {'z0var':z0var, 'var':var, 'theta':theta[var][0]}
            z0syntax += ["compute %(z0var)s = (%(var)s < %(theta)s) ." % subst]
        z0syntax += ["execute ."]

        spss.Submit(z0syntax)

        ## the bias correction is the proportion of bootstrap values less than the original
        ## simply compute the mean of the indicators

        z0 = Descriptives(z0rows, ["Mean"], variableLabels=z0vars)
        z0.Submit()
        z0.createDataset("z0_", activate=True)

        # IDF.Normal will cause an error if 0 or 1 is ever encountered
        # set these to somewhat arbitrary values to avoid this problem
        p0 = 1/(2.0*self.iterations)
        # at last, a list comprehension using dictionary substitution
        z  = ["if (%(var)s <= 0) %(var)s = %(min_p)f ." % {'var':z0var, 'min_p':p0} for z0var in z0vars]
        p0 = 1 - p0
        z += ["if (%(var)s >= 1) %(var)s = %(max_p)f ." % {'var':z0var, 'max_p':p0} for z0var in z0vars]
        z += ["compute %(var)s = IDF.Normal(%(var)s, 0, 1) ." % {'var':z0var} for z0var in z0vars]
        z += ["execute ."]
        spss.Submit(z)

        z0 = Descriptives(z0vars)
        z0.Submit()
        z0 = z0.createDataset(bc_dataset, activate=True)
        #print "---------- Bias-Corrections z0 ----------"
        #print z0.variables
        #print z0.data
        
        #
        # COMPUTE THE ACCELERATION, a
        #
        # These require jackknifed (leave-one-out) statistics for each case in the original data
        #

        # recall the keys we used to request the values, long ago
        #keys = oms.getKeys()
        ## the same keys are stored as bootrep.variables...
        keys = dataset.variables

        spss.Submit(["dataset activate %s ." % self.completeDataset])
        jack = Job(self.syntax, self.getOMS(), sample=JackKnife())
        jackData = jack.Submit()
        #print "JackKnife replicates:\n", "\n".join(jackData)

        # send the jackknifed means to a dataset
        jackrep = jack.createDataset("jackknife_replicates", activate=True)
        # switch back to the jackknife replicates
        #print "activate the jackknife dataset:", jackrep.data
        jackrep.activate()
        N = spss.GetCaseCount()

        skew = Descriptives(keys, statistics=["Skewness"])
        skewData = skew.Submit()
        #print "Skewness: ", skewData
        accel = skew.createDataset("acceleration", activate=True)

        # we're really computing the following adjustment to the skewness:
        # multiplier = -sqrt(N-1)*(N-2)/(6*N)

        adjust = "compute %(var)s = %(multiplier)f * %(var)s ."
        syntax = []
        multiplier = accelerationMultiplier(N)
        for var in keys:
            syntax += [adjust % {'var':var, 'multiplier':multiplier}]
        syntax += ["execute ."]
        spss.Submit(syntax)

        # calculate percentiles of the normal distribution
        zLabels = {"zlower":"z(alpha)", "zupper":"z(1-alpha)"}
        zalpha = Dataset("zalpha", varNames=["zlower","zupper"], varLabels=zLabels)
        zalpha.addRow(["0.025","0.975"])
        zalpha.Submit()
        zalpha.activate()
        #print "\n".join(zalpha.commands())
        #print "--- z values ---"
        #print zalpha.data

        syntax  = ["compute %(var)s = IDF.Normal(%(var)s, 0, 1) ." % {'var':var} for var in zalpha.variables]
        syntax += ["execute ."]
        spss.Submit(syntax)

        match = "match files file=%s ." % "\n  /file=".join([bc_dataset, "acceleration", "zalpha"])
        spss.Submit(match)
        zvars = jackrep.variables + z0.variables + zalpha.variables
        zdict = {}
        for zvar in zvars:
            zdict[zvar] = zvar
        zdict.update(zLabels)
        # this may seem silly, but we ask for the mean of a dataset with one row...
        # ... because that lets us retrieve the values in an XMLWorkspace
        zjob = Descriptives(zvars, ["Mean"], zdict)
        zdata = zjob.Submit()

        # once upon a time we saved z0key; such foresight

        zlower = float(zdata["zlower"][0])
        zupper = float(zdata["zupper"][0])
        alpha_lower = {}
        alpha_upper = {}
        for var in jackrep.variables:
            bias = float(zdata[z0key[var]][0])
            a = float(zdata[var][0])
            #print "Variable %s : Bias z0: %s Acceleration a: %s" % (var, bias, a)
            bz = bias + zlower
            alpha_lower[var] = [bias + (bz/(1 - a*bz))]
            bz = bias + zupper
            alpha_upper[var] = [bias + (bz/(1 - a*bz))]

        # next we need to compute CDF.Normal(alpha, 0, 1) for these

        # save all the values in a dataset with only one row
        # even though everything is named alpha, they are really not in [0,1] yet
        alphaNames = []
        alphaNamesLower = {}
        alphaNamesUpper = {}
        alphaData = {}
        for var in jackrep.variables:
            # first work on the lower limit
            name = "%s_alpha_lower" % var
            alphaNames += [name]
            alphaNamesLower[var] = name
            alphaData[name] = alpha_lower[var]
            # now work on the upper limit
            name = "%s_alpha_upper" % var
            alphaNames += [name]
            alphaNamesUpper[var] = name
            alphaData[name] = alpha_upper[var]

        alpha = Dataset("alpha", alphaNames)
        alpha.addDataFromDictionary(alphaData)
        alpha.Submit()
        alpha.activate()
        # at last we deserve to call them alpha
        alphaTransform = ["compute %(alpha)s = CDF.Normal(%(alpha)s, 0, 1) ." % {'alpha':name} for name in alphaNames]
        alphaTransform += ["execute ."]
        spss.Submit(alphaTransform)

        alphaJob = Descriptives(alphaNames)
        alphaValues = alphaJob.Submit()

        #print "alpha values... ", "\n".join([key+": "+" ".join(alphaValues[key]) for key in alphaValues.iterkeys()])

        # need to go variable by variable; each one needs its own alpha_lower and alpha_upper
        dataset.activate()
        percentiles="""FREQUENCIES
           VARIABLES=%s
          /PERCENTILES=%10.8f %10.8f
          /ORDER=ANALYSIS
          /FORMAT=NOTABLE ."""

        cidata = {}
        for var in ("theta", "parameter", "alpha_lower", "alpha_upper", "ci_lower", "ci_upper"):
            cidata[var] = []
        bvarDict = VariableDict()

        cilabel = {}
        for var in dataset.variables:
            lower = 100.0*float(alphaValues[alphaNamesLower[var]][0])
            upper = 100.0*float(alphaValues[alphaNamesUpper[var]][0])
            #print "percentiles: ", lower, upper
            ptiles = percentiles % (var, lower, upper)    # .split("\n")
            
            cioms = OMS(commands="Frequencies", subtypes="Statistics")  #, columnDimensions="Variables")
            cioms.requestGroup("ptile","ptile_value")
            cioms.Submit()
            #print ptiles
            spss.Submit(ptiles)
            cioms.end()
            ci = cioms.getValues()
            cilabel[bvarDict.VariableIndex(var)] = var
            cidata["parameter"] += [bvarDict.VariableIndex(var)] # should give in and allow string variable types

            cidata["theta"] += [theta[var][0]]
            
            # utility function to deal with ugly data structure
            def setCI(cidata, stat, ci, civalue, index):
                try:
                    cidata[stat] += [ci[civalue][index]]
                except:
                    cidata[stat] += ["."]

            setCI(cidata,"ci_lower", ci, "ptile_value", 0)
            setCI(cidata,"ci_upper", ci, "ptile_value", 1)            
            setCI(cidata,"alpha_lower", ci, "ptile", 0)
            setCI(cidata,"alpha_upper", ci, "ptile", 1)

        ## try block needs to surround EACH of these to deal with missing values            
        ##    cidata["alpha_lower"]+= [ci["ptile"][0]]
        ##    cidata["alpha_upper"]+= [ci["ptile"][1]]
        ##    cidata["ci_lower"] += [ci["ptile_value"][0]]
        ##    cidata["ci_upper"] += [ci["ptile_value"][1]]
        ####    for var in cidata.iterkeys():
        ####        print "CI:  %s %s %s %s %s" % (cidata["parameter"], cidata["alpha_lower"], cidata["alpha_upper"], cidata["ci_lower"], cidata["ci_upper"])
            #print "confidence intervals", ci
            #print "ptile:       ", ci["ptile"]
            #print "ptile value: ", ci["ptile_value"]
            cioms.deleteXMLWorkspace()

        valLabels = {0:"Original Statistic",  1:"BCa (lower)", 2:"BCa (upper)", 3:"alpha (lower)", 4:"alpha (upper)"}
        civars = ["theta", "ci_lower", "ci_upper", "alpha_lower", "alpha_upper"]
        civarDict = {}
        for i in range(len(civars)):
            civarDict[civars[i]] = i
        cidata_table = {"theta":[], "parameter":[], "statistic":[], "value":[]}
        for v in range(len(cidata["parameter"])):
            for varindex in range(len(civars)):
                var = civars[varindex]
                cidata_table["parameter"] += [v]
                cidata_table["statistic"] += [varindex]
                cidata_table["value"] += [cidata[var][v]]

        #print cidata_table


        #print "Recall theta: ", theta
        #print "confidence interval data:"
        #print cidata
        #print cilabel
        cidataset = Dataset("cidata", ["parameter", "statistic", "value"])
        cidataset.addDataFromDictionary(cidata_table)
        cidataset.Submit()
        cidataset.activate()
        vallabelsyntax  = ["value labels parameter %s ." % " ".join(["%s '%s'" % (i, cilabel[i]) for i in cilabel.iterkeys()])]
        vallabelsyntax += ["value labels statistic %s ." % " ".join(["%s '%s'" % (val, valLabels[val]) for val in valLabels])]
        vallabelsyntax += ["variable level parameter statistic (nominal) ."]
        #print "Label syntax:\n", vallabelsyntax
        spss.Submit(vallabelsyntax)
        self.bcaDataset = cidataset
        return cidataset

    def bcaSummary(self, useCTables=False):
        if useCTables:
            self.bcaDataset.summaryCTableBy("parameter", "statistic", ["value"], "BCa Confidence Intervals", stats="Mean")
        else:
            self.bcaDataset.summaryTableBy("parameter", "statistic", ["value"], "BCa Confidence Intervals", stats="Mean")


####
#### THE VariableDict object IS STOLEN FROM spssaux.py
#### only to solve an import problem
####
##
##class VariableDict:
##    """A Python dictionary indexed by variable name for use with
##    the SPSS Variable Dictionary.  The Python dictionary
##    does not contain SPSS dictionary properties other than name and index.  Other
##    variable properties are retrieved dynamically when requested.
##    
##    If no argument is supplied, all variable names in the SPSS dictionary are indexed.
##    If a list of names or indexes is supplied, only those variables are included.
##    If safe == True, each time a variable property is retrieved, the actual
##    SPSS dictionary is checked to see if the index still refers to that variable,
##    and KeyError is raised if it is not.
##    """
##    def __init__(self, namelist=None, safe=False):
##            allnames = [spss.GetVariableName(v) for v in range(spss.GetVariableCount())]
##            if namelist is None:
##                    self.vdict = dict([(n, index) for index, n in enumerate(allnames)])
##            elif type(namelist[0]) is str:		# got list of names
##                    self.vdict = dict([(n, index) for index, n in enumerate(allnames) if n in namelist])
##            else:								# got list of indexes
##                    self.vdict = dict([(n, index) for index, n in enumerate(allnames) if index in namelist])
##            self.safe = safe
##            self.numvars = len(self.vdict)
##
##    def __getitem__(self, index):
##        return self.vdict[index]
##    
##    def VariableCount(self):
##            "return the number of variable in this dictionary"
##            return self.numvars
##            
##    def VariableIndex(self,id):
##            "return the index of the variable with name or index id."
##            # if varDict is safe, check that the named variable is still at the saved index value
##            if type(id) is str:
##                    res = self.vdict[id]
##                    if self.safe and id != spss.GetVariableName(res): raise KeyError
##            else:
##                    res = id
##            
##            return res
##
##    def VariableName(self, id=None):
##            "return the name of the variable with name or index id."
##            # admittedly pointless if id is in fact a name, but allows the class to support names and
##            # indexes uniformly
##            if id is None:
##                    return None
##            return spss.GetVariableName(self.VariableIndex(id))
##
##    def VariableLabel(self, id):
##            "return the label of the variable with name or index id."
##            return spss.GetVariableLabel(self.VariableIndex(id))
##            
##    def VariableLevel(self, id):
##            "return the measurement level of the variable with name or index id."
##            return spss.GetVariableMeasurementLevel(self.VariableIndex(id))
##
##    def VariableType(self, id):
##            """return the type of the variable with name or index id.
##            0 = numeric
##            >0 = string of that length
##            """
##            return spss.GetVariableType(self.VariableIndex(id))
##        
##    def VariableFormat(self, id):
##            "return the format of the variable with name or index id."
##            return spss.GetVariableFormat(self.VariableIndex(id))
##
##    def	ValueLabels(self, id):
##            "return the set of value labels of the variable with name or index id as a dictionary."
##            return GetValueLabels(self.VariableIndex(id))
##
##    def Attributes(self, id=None):
##            """return a dictionary of attributes for variable or index id.  If id = None
##            the datafile attributes are returned
##            """
##            return GetAttributesDict(self.VariableName(id))
##
##    def MissingValues(self, id):
##            "return a string listing the missing value codes for variable or index id"
##            return GetMissingValues(self.VariableIndex(id))
##    
##    def VariableNamesFromLabel(self, label):
##            """return a list of variable names for variables that have as label the specified text.
##            This may be useful when SPSS output returns a variable label when the name is neede.  For
##            example, getSHOW('weight') returns the label of the weight variable.  Variable labels need
##            not be unique, however.
##            """
##            return [name for (name, index) in self.vdict.items() if spss.GetVariableLabel(index) == label]
##
