"""Calculate and display Pareto chart with acceptance intervals computed using Monte Carlo simulation.
Requires at least SPSS 15.0.1 and corresponding versions of spssaux and spssdata modules."""

# copyright(c) SPSS Inc, 2007
# history
# 2007-jan-22 initial version

__author__  =  'spss'
__version__ =  '1.0.0'


# translatable messages
_msg1 = "Count"
_msg2 = "Acceptance intervals computed for alpha ="

import random, spss, spssaux, spssdata

def chart(varname, title=None, closeds=True, alpha=.05, totalcategories=None,
          template=None):
    """Produce Pareto chart for variable varname including acceptance intervals.
    
    This module requires at least SPSS 15.0.1 and corresponding versions of spssaux and spssdata.
    
    varname is the variable to chart.
    title is an optional title string
    closeds controls whether the calculated dataset is discarded or left open, defaulting to discarding.
    alpha is the confidence interval level, defauting to .05.
    totalcategories can be set to account for potential categories that did not appear in the data
    when calculating the ci's.  By default these are based on the number of categories found in the data, but they
    can be overridden by specifying a different total number of categories.
    template can be supplied as a filespec for a chart template to be applied."""
    
    gpldict = {}   # parameters for the graph GPL
    gpldict["varname"] = varname
    gpldict["msg1"] = _msg1
    gpldict["msg2"] = _msg2
    vard = spssaux.VariableDict(namelist=varname)
    varlabel = vard[0].VariableLabel
    varlabel = varlabel.replace('"', r'\"')
    if not varlabel:
        varlabel = varname
    gpldict["varlabel"] = varlabel
    gpldict["alpha"] = alpha
    
    # first aggregate variable into new dataset.  Ensure current dataset has a name
    dsname = spssaux.GetActiveDatasetName()
    if dsname is None:
        dsname = "T" + str(random.random())
        spss.Submit("DATASET NAME " + dsname)
    aggdsname = "T" + str(random.random())
    spss.Submit("""
    DATASET DECLARE  %(aggdsname)s.
    AGGREGATE /OUTFILE= '%(aggdsname)s' /BREAK=%(varname)s /counts_ = N.
    DATASET ACTIVATE %(aggdsname)s.
    SORT CASES BY counts_(D).""" % {"aggdsname":aggdsname, "varname": varname})
    
    curs = spssdata.Spssdata(accessType='w', names=False)
    data = curs.fetchall()
    ncats = len(data)
    totalcases = sum(ct for i, ct in data)
    
    # calculate acceptance intervals and add to dataset
    if totalcategories is None:
        virtualcats = ncats
    else:
        virtualcats = totalcategories
    p = ParetoAcceptance(virtualcats, totalcases)
    p.simulateUntilConvergence()
    
    # write calculated interval endpoints to aggregate dataset
    
    curs.restart()
    curs.append(spssdata.vdef("lowerLimit",vlabel="Lower Limit", vfmt=("F", 8,0)))
    curs.append(spssdata.vdef("upperLimit",vlabel="Upper Limit", vfmt=("F", 8,0)))
    curs.commitdict()
    for i in range(ncats):
        curs.fetchone()
        curs.casevalues(p.acceptanceInterval[i])
    curs.CClose()
    if title:
        gpltitle = ' GUIDE: text.title(label("%s"))' % title.replace('"', r'\"')  #GPL uses \ for escaping "
    else:
        gpltitle = ""
    gpldict["gpltitle"] = gpltitle
    if template:
        template = 'TEMPLATE=[' + spssaux._smartquote(template) + ']'
    else:
        template = ''
    gpldict["template"] = template
    
    spss.Submit(r"""GGRAPH
  /GRAPHDATASET NAME="graphdataset" VARIABLES=%(varname)s upperLimit lowerLimit
  counts_   /GRAPHSPEC SOURCE=INLINE %(template)s.
  BEGIN GPL
   SOURCE: s=userSource(id("graphdataset"))
   DATA: %(varname)s=col(source(s), name("%(varname)s"), unit.category())
   DATA: upperLimit=col(source(s), name("upperLimit"))
   DATA: lowerLimit=col(source(s), name("lowerLimit"))
   DATA: counts_=col(source(s), name("counts_"))
   GUIDE: axis(dim(1), label("%(varlabel)s"))
   GUIDE: axis(dim(2), label("%(msg1)s"))
   %(gpltitle)s
    GUIDE: text.footnote(label("%(msg2)s %(alpha).3f"))
   SCALE: cat(dim(1), sort.data())
   SCALE: linear(dim(2), include(0))
   ELEMENT: interval(position(region.spread.range(%(varname)s*(
    lowerLimit+upperLimit))), shape(shape.square), transparency(transparency."1.0"))
   ELEMENT: point(position(%(varname)s*counts_), shape(shape.circle),size(size."6pt"),
   transparency(transparency."0.0"))
   END GPL.""" % gpldict)

    # reactivate original dataset
    
    spss.Submit("DATASET ACTIVATE %s." % dsname)
    if closeds:
        spss.Submit("DATASET CLOSE %s." % aggdsname)

# simulation class

class ParetoAcceptance(object):
    """
    Simulation to obtain acceptance intervals for Pareto-Wilkinson chart

    Wilkinson, Leland. "Revising the Pareto Chart".
    The American Statistician, November 2006, Vol. 60, No. 4 pp. 332-334
    
    Typical usage:

    pareto = ParetoAcceptance(bins=10, objects=100)
    pareto.simulateUntilConvergence(200)  
    acceptanceIntervals = pareto.simulateUntilConvergence()

    # acceptanceIntervals will be a list of tuples (lower, upper) for each bin    
    """
    def __init__(self, bins, objects):
        """
        bins    = number of categories
        objects = number of objects to be placed in the bins
        """
        self.bins = bins
        self.objects = int(objects)
        self.frequencies = [{} for bin in xrange(bins)]
        self.total_trials = 0
        self.acceptanceInterval = []

    def simulate(self, trials=1000):
        """
        Simulate the order statistics, placing objects into bins with equal probability.
        """
        bins1 = self.bins - 1
        self.total_trials += trials
        for i in xrange(trials):
            sample = [0]*self.bins
            for obj in xrange(self.objects):
                sample[random.randint(0, bins1)] += 1
            sample.sort(reverse=True)
            for freq, c in zip(self.frequencies, sample):
                freq[c] = freq.get(c,0) + 1        

    def acceptanceIntervals(self, alpha=0.05):
        """
        Calculate the alpha/2 and 1-alpha/2 percentile for each bin.

        The result is a list containing a tuple (lower, upper) for each bin. 
        The most recently calculated intervals are stored as self.acceptanceInterval        
        """
        accept = []
        n = self.total_trials
        ptile = int(n*(alpha/2.0))
        for freq in self.frequencies:
            lower = self._getPercentile(freq, ptile)
            upper = self._getPercentile(freq, ptile, reverse=True)
            accept.append((lower, upper))
        self.acceptanceInterval = accept
        return accept

    def _getPercentile(self, freq, ptile, reverse=False):
        count = freq.keys()
        count.sort(reverse=reverse)
        ccum = 0
        for c in count:
            ccum += freq[c]
            if ptile <= ccum:
                return c
        # this should be impossible to reach
        return -1

    def simulateUntilConvergence(self, trialsPerBatch=200, converge=5, maximumTrials=100000):
        """
        Add a batch of trials and compute acceptance intervals until successive intervals are identical.
        
        trialsPerBatch defaults to 200; after each batch, acceptance intervals are computed.
        Convergence is declared when the acceptance intervals are unchanged for a run of consecutive batches.
        The length of the run is controlled by the converge parameter.
        maximumTrials is used to ensure the simulation does not continue for an unacceptably long time.
        """
        trials = 0
        run = 0
        while run < converge and trials < maximumTrials:
            trials += trialsPerBatch
            self.simulate(trialsPerBatch)
            previous = self.acceptanceInterval
            current = self.acceptanceIntervals()
            if previous == current:
                run +=1
            else:
                run = 0
        return self.acceptanceInterval
        
    def report(self):
        """
        Placeholder for a more substantial report, if needed.
        """
        for interval in self.acceptanceInterval:
            print interval

if __name__ == "__main__":
    trials = 200
    bins = 10
    objects = 100
    print "Acceptance Intervals with %s bins, %s objects, %s trials" % (bins, objects, trials)
    pareto = ParetoAcceptance(bins, objects)
    pareto.simulateUntilConvergence(200)
    pareto.report()
    print "goodbye"
    

        

 
