
"""
module tr_rnd

This is an SPSS/Python client for Mads Haahr's random number generator at www.random.org

It depends on the module randomwrapy, which provides a simple Python wrapper to the REST interface
at www.random.org.

Documentation, including installation instructions can be found in tr_rndReadme.

Both randomwrapy and tr_rndReadme should be included with this file.

 This client by George Dunbar, University of Warwick (Copyright George Dunbar, 2008)
 It is distributed under the Gnu General Public License.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    See <http://www.gnu.org/licenses/> for a copy of the GNU General Public License.
    For use that falls outside this license, please contact me.

    May 22nd 2008    

    """

__author__  =  'George Dunbar'
__version__ =  '0.1'
version = __version__    

def trandi(min=0, max=0, howmany=0, replacement=None, groups=0, varname='rndno', rdataset='*'):
    """Request random integers from a true random number generator and place them in an SPSS variable.
    
    min is the lowest number in the range
    max the biggest
      usage example
        trandi(1, 15)
      will randomly order the integers 1-15
    limit on min and max is 1*10E9 (+/-), but max - min =< 10E4
    
    howmany is howmany numbers you want; normally only specified if there is replacement ie. replacement is not None
      usage example
        trandi(0, 100, 20)
         will randomly pick 20 integers in the range [0, 100]; they will all be different (limitation: howmany must be =< range)
        trandi(0, 100, 20, replacement=1)
         as before, but there can be duplicates; this is the more usual form when howmany is specified
        trandi(1,6, 12, replacement=1)
         roll a dice 12 times
    limit on howmany is 10E4
    
    if replacement is None or 0, there is no replacement, otherwise there is replacement

    groups, if specified, should be an integer > 1, and will flatten the numbers into the range [0, groups-1]
      min should be set to 0, and groups should be a divisor of max (or the groups will not be the same size)
      
    varname will be the variable name the numbers are placed into, in the spss file
      if the default name is accepted, the variable name will be "rnd" + (index of last var + 1)
      
    rdataset if '*' will create the data in the active data set, if None it goes in a fresh spss dataset (new editor window)

    """
    import spss
    import randomwrapy
    if min == 'zero': min = 0
    if max - min < howmany and replacement == None or replacement == 0:
        raise Exception, "*** Warning: when sampling without replacement, the number of values requested cannot be greater than max - min. To sample with replacement, add 'replacement=1' as an additional argument. ***"
    if min >= max:
        raise Exception, "*** Warning: the minimum value must be smaller than the maximum. ***"
    #fetch the numbers
    if replacement == None or replacement == 0:
        numbers = randomwrapy.rnumlistwithoutreplacement(min, max)
        if howmany: numbers = numbers[:howmany]
    else:
        print "how", howmany, "max, min", max, min
        numbers = randomwrapy.rnumlistwithreplacement(howmany, max, min)
    numbers = map(int, numbers)
    if groups > 1:
        numbers = map(lambda x: x%groups, numbers)
        if min > 0: print "*** Warning: min is not set to zero. When requesting 'grouped' data, min must be zero to allow equal size groups ***"
    
    #put them into spss (this try / except is taken directly from SPSS documentation)
    try:
        spss.StartDataStep()
    except spss.SpssError:
        spss.Submit("Execute")
        spss.StartDataStep()
    rdata = spss.Dataset(name=rdataset)
    if rdataset == '*':
        numvars = len(rdata.varlist)
    else:
        numvars = 0
    if varname == 'rndno': varname = 'rnd' + str(numvars)
    rdata.varlist.append(name=varname, type=0)
    caselist = rdata.cases
    if numvars == 0 or len(caselist) < len(numbers):
        for case in range(0, len(numbers) - len(caselist)):
            caselist.append([None])
    targ= len(numbers)
    caselist[0:targ, numvars] = numbers
    spss.EndDataStep()


def tsamp(existingvar, newvar, percent=None, lilN=None, truncateAt=None, placeBeside=0):
    """Create a variable 'newvar' that has the value 1 for 'percent' of cases in 'existingvar', and 0 for the rest.

    'existingvar' must be present in the active dataset.
    However, it is an error if a variable called 'newvar' already exists.
    Both 'existingvar' and 'newvar' should be passed to the function as strings (i.e. passed in quotes).
    percent of cases is rounded to the nearest whole number. E.g. 10% of 11 cases will be 1 case, and 50% of 15 is 7 cases.
    lilN is an alternative to percent. If supplied, it specifies how many cases to sample, and should be an integer.
    
    usage example
    tsamp('myvar','myfilt',40)
        This will create a new variable 'myfilt' in the active dataset.
        It will contain the integer 1 for 40% of the cases in 'myvar'. In practice, this will be 40% of the cases in the dataset.
        tsamp('myvar','myfilt', percent=40) is equivalent.

    tsamp('myvar','myfilt', lilN=40)
        This will create a filter variable, 'myfilt', that selects 40 cases.
        
    truncateAt can be set to an integer, m. If it is, then only the first m cases are considered when sampling (where 1 is the first case).
    If placeBeside = 1, then the new variable is inserted beside the existing one. Otherwise, it is placed at the end,
    after the last variable in the dataset.

    If there are cases with missing data, they have the same probability of being sampled as complete cases.
    If there are trailing missing values in 'existingvar' only because there are other variables in the dataset with more cases,
    then truncateAt may be a way to trim them. For example, if you wanted to select a sample of cases from a variable called
    'myvar' that had five cases, but there was another variable in the dataset with 100 cases, 'myvar' would appear also
    to have a hundred cases, and a 40% sample would have 40 cases, many of them, of course, corresponding to missing values in 'myvar'.
    truncateAt lets you restrict sampling to the first m cases.
    
    usage example
    tsamp('myvar','myfilt', 40, truncateAt=6, placeBeside=1)
        This will create a new variable 'myfilt' in the active dataset, placed immediately to the right of 'myvar'.
        The new variable will contain the integer 1 for 40% of the first six cases in 'myvar'.

    If the supplied parameter lilN > truncateAt, all cases up to truncateAt are selected. For example:
    tsamp('rnd0','myfrt',lilN=9, truncateAt=8)
        This will select the first 8 cases.
    """
    import spss
    import randomwrapy
    
    if type(truncateAt) == type(1):
        bigN = truncateAt
    else:
        for i in range(spss.GetVariableCount()):
            if spss.GetVariableName(i) == existingvar:
                index = [i] # index of existingvar, could do more easily using spssdata which allows reference by variable name
                break
        dataCursor = spss.Cursor(index)
        vardata = dataCursor.fetchall() # this includes trailing missing values, so it would be simpler just to ask for the number of cases in the file
        dataCursor.close()
        bigN = len(vardata)
        
    if lilN == None:
        prop = float(percent)/100
        lilN = int(prop * bigN)
    
    numbers = randomwrapy.rnumlistwithoutreplacement(0, bigN-1) #minus 1 here to align numbers with case numbering in Data window
    numbers = map(int, numbers[:lilN])
    newdata = buildfilterlist(bigN, numbers)

    try:
        spss.StartDataStep()
    except spss.SpssError:
        spss.Submit("Execute")
        spss.StartDataStep()

    rdata = spss.Dataset(name='*')
    caselist = rdata.cases
    targ = len(newdata)
    
    if placeBeside == 1:
        oldvar = rdata.varlist[existingvar]
        outdex = oldvar.index + 1
        rdata.varlist.insert(name=newvar, type=0, index=outdex)
        caselist[0:targ, outdex] = newdata
    else:
        numvars = len(rdata.varlist)
        rdata.varlist.append(name=newvar, type=0)
        caselist[0:targ, numvars] = newdata
        
    spss.EndDataStep()


def applyfilter(index, item, idents):
    if index in idents:
        return 1
    else:
        return 0
        
def buildfilterlist(length, identifiers):
    flist = [0 for x in range(0, length)]
    flist = [applyfilter(indx, x, identifiers) for indx, x in enumerate(flist)]
    return flist

def reportquota():
    import urllib2
    request = urllib2.Request("http://www.random.org/quota/?format=plain")
    request.add_header('User-Agent', 'randomwrapy/0.1 very alpha')
    opener = urllib2.build_opener()
    quota = opener.open(request).read()
    print "This IP address has", quota, "bits left. Visit http://www.random.org/quota for more information."